"""Plot utilities to visualize statistics

"""
from typing import Sequence, TypeVar

T = TypeVar('T')


def cycle_colors(
    cm: Sequence[T],
    length: int
) -> Sequence[T]:
    """Cycle colors

    cm : sequence of colors
    length : int
        Length you want

    """
    len_cm = len(cm)
    if length <= len_cm:
        return cm[:length]
    colors = []
    cycle = -1
    for i in range(length):
        if i % len_cm == 0:
            cycle += 1
        if i >= len_cm:
            i -= len_cm * cycle
        colors.append(cm[i])
    return colors


def to_hex_color(x):
    """Ignore alpha"""
    if isinstance(x[0], float):
        x = [int(255*_x) for _x in x]
    return '#{:02X}{:02X}{:02X}'.format(x[0], x[1], x[2])



import json
import os.path
from functools import cached_property
from pathlib import Path
from typing import List, Optional, Sequence, Union

import albumentations
import numpy as np

from ..base import Dataset
from ..utils import imread_asarray


class BBBC041(Dataset):
    """P. vivax (malaria) infected human blood smears [1]_

    Images are in .png or .jpg format. There are 3 sets of images consisting of
    1364 images (~80,000 cells) with different researchers having prepared each
    one: from Brazil (Stefanie Lopes), from Southeast Asia (Benoit Malleret),
    and time course (Gabriel Rangel). Blood smears were stained with Giemsa
    reagent.

    These images were contributed by Jane Hung of MIT and the Broad Institute in
    Cambridge, MA. [1]_

    There is also a Github reposity that lists malaria parasite imaging datasets
    (blood smears) [2]_.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ```transforms``` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    training : bool, default: True
        Load training set if True, else load testing one

    Notes
    -----
    - Label categories: all 7 cats, ['difficult', 'gametocyte', 'leukocyte',
      'red blood cell', 'ring', 'schizont', 'trophozoite']
    - 1208/120 training/test split. So not 1368 images as written in the
      description.
    - png and jpg extension; training images are in RGB space in PNG format,
      while test images are in YUV space in JPEG format.
    - YUV will be automatically detected when read and cast to RGB
    - Two resolutions; depending on training/test: (1200, 1600) for training,
      (1383, 1944) for test

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC041
    .. [2] https://github.com/tobsecret/Awesome_Malaria_Parasite_Imaging_Datasets

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """

    # Dataset's acronym
    acronym = 'BBBC041'

    def __init__(
        self,
        root_dir: str,
        *,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # specific to this dataset
        training: bool = True,
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this one here
        self.training = training

    def get_image(self, p: Path) -> np.ndarray:
        img = imread_asarray(p)
        return img

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        file_list = []
        for name in self.ids:
            # Dont' know why they put '/' at front in json
            file_list.append(root_dir / name.lstrip('/'))
        return file_list

    @cached_property
    def ids(self) -> list:
        return [d['image']['pathname'] for d in self.metadata]

    @cached_property
    def metadata(self) -> List[dict]:
        name = 'training.json' if self.training else 'test.json'
        p = self.root_dir / name
        with open(p, 'r') as f:
            data = json.load(f)
        return data



from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional

import albumentations
import cv2
import numpy as np
import scipy.ndimage as ndi
import tifffile
from PIL import Image

from ..base import MaskDataset


class BBBC039(MaskDataset):
    """Nuclei of U2OS cells in a chemical screen [1]_

    This data set has a total of 200 fields of view of nuclei captured with
    fluorescence microscopy using the Hoechst stain. These images are a sample
    of the larger BBBC022 chemical screen. The images are stored as TIFF files
    with 520x696 pixels at 16 bits.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    training : bool, default: True
        Load training set if True, else load testing one

    Notes
    -----
    - Split (training/valiadation/test)
        - `training=True` combines 'training' with 'validation'
    - Annotate objs not touching each other with 1 and use 2, 3, ... for the
      touching ones. It is great and clever, but it does not follow the form of
      other instance segmented masks. ``get_mask()`` will make a instance
      labeled mask (each obj has unique labels). After labeling max label is 231
      for training, and 202 for test. So having masks of dtype UINT8 is fine.
    - Max label is 3 (in original annotation)
    - Sample of larger BBBC022 and did manual segmentation
    - Possible overlap some with DSB2018
    - Mask is png but (instance) value is only stored in RED channel
    - Maximum value is 2**12

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC039

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """

    # Dataset's acronym
    acronym = 'BBBC039'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        # specific to this dataset
        training: bool = True,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        # specific to this dataset
        self.training = training

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        img = (img / 2**4).astype(np.uint8)
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        return img

    def get_mask(self, p: Path) -> np.ndarray:
        mask = np.asarray(Image.open(p))[..., 0]
        max_val = mask.max()
        if max_val == 0:
            return mask
        else:
            inst_mask, n_labels = ndi.label(mask == 1, output='uint8')  # int32 by default
            for m in range(2, max_val+1):
                labeled, n = ndi.label(mask == m, output='uint8')
                inst_mask += np.where(labeled == 0, 0, labeled + n_labels)
                n_labels += n
        return inst_mask

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = root_dir / 'images'
        file_list = []
        for name in self.ids:
            p = parent / name
            file_list.append(p.with_suffix('.tif'))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        parent = root_dir / 'masks'
        anno_list = []
        for name in self.ids:
            p = parent / name
            anno_list.append(p)
        return dict((k, v) for k, v in enumerate(anno_list))

    @cached_property
    def ids(self) -> list:
        def _readlines(path):
            with open(path, 'r') as f:
                lines = f.readlines()
            return list(map(lambda s: s.strip(), lines))
        meta_dir = self.root_dir / 'metadata'
        if self.training:
            # Combine training and validation
            meta_file = meta_dir / 'training.txt'
            _ids = _readlines(meta_file)
            meta_file = meta_dir / 'validation.txt'
            _ids += _readlines(meta_file)
        else:
            meta_file = meta_dir / 'test.txt'
            _ids = _readlines(meta_file)
        return _ids



"""Define a base class and its interface

``Dataset`` is the base of all datasets

``MaskDataset`` is the base of datasets that have mask annotation
"""

import abc
import random
from pathlib import Path
from typing import Any, Dict, Iterator, Optional, Sequence, Union

import albumentations
import cv2
import numpy as np


class DatasetInterface(metaclass=abc.ABCMeta):
    """Dataset interface

    Attributes
    ----------
    __repr__
    __len__
    __getitem__
    acronym
    root_dir
    file_list

    Methods
    -------
    get_image

    """

    @abc.abstractmethod
    def __repr__(self):  # common
        """Print info of dataset"""
        ...

    @abc.abstractmethod
    def __len__(self):  # common
        ...

    @abc.abstractmethod
    def __getitem__(self, ind):  # common
        """Given index returns a dictionary of key(s) and array"""
        ...

    @abc.abstractmethod
    def get_image(self, key):  # required
        """Get an image"""
        ...

    @classmethod
    @property
    @abc.abstractmethod
    def acronym(cls):  # required
        """Assign acroym for a subclass dataset"""
        ...

    @property
    @abc.abstractmethod
    def root_dir(self):  # required
        """Path to root directory"""
        ...

    @property
    def file_list(self):  # required
        """A list of pathes to image files"""
        ...


class Dataset(DatasetInterface):
    """Base to define common attributes and methods for [`MaskDataset`, ...]

    Attributes
    ----------
    __repr__
    __len__
    __iter__
    root_dir
    output
    transforms
    num_calls
    grayscale : optional
    grayscale_mode : optional
    num_channels : optional

    Methods
    -------
    __getitem__
    _drop_missing_pairs
    to_gray

    Notes
    -----
    Required attributes in subclass
        - ``anno_dict``
        - ``__getitem__()``
        - ``get_image()``

    """

    def __repr__(self):
        return self.acronym

    def __len__(self):
        """Length of dataset. Can be overwritten with ``num_calls``"""
        if self.num_calls is not None:
            return self.num_calls
        return len(self.file_list)

    def __iter__(self):
        return IterDataset(self)

    @property
    def root_dir(self) -> Path:
        if hasattr(self, '_root_dir'):
            _root_dir = getattr(self, '_root_dir')
            if not isinstance(_root_dir, Path):
                return Path(_root_dir)
            return _root_dir
        raise NotImplementedError("Attr `_root_dir` not defined")

    @property
    def output(self) -> str:
        """Determine return(s) when called, fixed to 'image'"""
        return 'image'

    @property
    def transforms(self) -> Optional[albumentations.Compose]:
        """Transform images and masks"""
        if hasattr(self, '_transforms'):
            return getattr(self, '_transforms')
        return None

    @property
    def num_calls(self) -> Optional[int]:
        """Number of calls that will override __len__"""
        if hasattr(self, '_num_calls'):
            return getattr(self, '_num_calls')
        return None

    @property
    def grayscale(self) -> Optional[bool]:
        """Flag for grayscale conversion"""
        if hasattr(self, '_grayscale'):
            return getattr(self, '_grayscale')
        return None

    @grayscale.setter
    def grayscale(self, val):
        self._grayscale = val

    @property
    def grayscale_mode(self) -> Optional[Union[str, Sequence[float]]]:
        """Determine grayscale mode one of {'cv2', 'equal', Sequence[float]}
        """
        if hasattr(self, '_grayscale_mode'):
            return getattr(self, '_grayscale_mode')
        return None

    @grayscale_mode.setter
    def grayscale_mode(self, val):
        self._grayscale_mode = val

    @property
    def num_channels(self) -> Optional[int]:
        """Number of image channels used for `to_gray()`"""
        if hasattr(self, '_num_channels'):
            return getattr(self, '_num_channels')
        return None

    def __getitem__(self, ind: int) -> Dict[str, np.ndarray]:
        """Get image

        Dataset does not any annotation available. It will only load 'image'.

        Parameters
        ----------
        ind : int
            Index to get path(s) from ``file_list`` attribute

        Attributes
        ----------
        self.file_list

        Other Parameters
        ----------------
        self._transforms
        self._num_calls
        self._grayscale
        self._grayscale_mode
        self._num_channels

        """
        # Randomize `ind` when `num_calls` set
        if self.num_calls is not None:
            if ind >= self.num_calls:
                raise IndexError('list index out of range')
            ind_max = len(self.file_list)
            ind = random.randrange(0, ind_max)
        # `output="image"`
        p = self.file_list[ind]
        image = self.get_image(p)
        if self.grayscale:
            num_channels = self.num_channels
            if num_channels is None:
                num_channels = len(p) if isinstance(p, list) else 3
            image = self.to_gray(
                image,
                grayscale_mode=self.grayscale_mode,
                num_channels=num_channels
            )
        if self.transforms:
            image = self.transforms(image=image)['image']
        return {'image': image}

    def _drop_missing_pairs(self) -> tuple:
        """Drop images and reindex the anno list (dict)

        Sometimes, not all images have annotation. For consistence, this func
        simply drops those images missing annotation.

        For example,
        - MurphyLab
        - BBBC018
        - BBBC020
        """
        file_list = getattr(self, 'file_list')
        anno_dict = getattr(self, 'anno_dict')
        _diff = set(range(len(file_list))).difference(set(anno_dict))
        diff = sorted(_diff)
        # logger.info(f'{self.acronym}:Dropping indices: {diff}')
        for i, ind in enumerate(diff):
            file_list.pop(ind-i)
        anno_dict = dict((i, v) for i, v in enumerate(anno_dict.values()))
        return file_list, anno_dict

    @staticmethod
    def to_gray(
        arr: np.ndarray,
        grayscale_mode: Optional[Union[str, Sequence[float]]] = None,
        num_channels: int = 3,
    ) -> np.ndarray:
        """Convert bioimage to grayscale

        Parameters
        ----------
        arr : image array
            Numpy image array whose shape is (h, w, 3)
        grayscale_mode : str or sequence of float, optional
            Choose a strategy for gray conversion. Three options are availble.
            Either one of {'cv2', 'equal'} or be a sequence of float numbers,
            which indicate linear weights of each channel.
        num_channels : int
            Explicitly set number of channels for `grayscale_mode='equal'`.
        """
        if isinstance(grayscale_mode, str):
            if grayscale_mode == 'cv2':
                if arr.shape[-1] != 3:
                    raise ValueError("Image arr should have RGB channels")
                arr = cv2.cvtColor(arr, cv2.COLOR_RGB2GRAY)
                arr = cv2.cvtColor(arr, cv2.COLOR_GRAY2RGB)
            elif grayscale_mode == 'equal':
                # Expect (h, w, ch) shape of array
                arr = (arr.sum(axis=-1) / num_channels).astype(arr.dtype)
                arr = cv2.cvtColor(arr, cv2.COLOR_GRAY2RGB)
            else:
                raise ValueError(f"Wrong `grayscale_mode={grayscale_mode}`")
        else:
            raise NotImplementedError("`grayscale_mode`")
        return arr


class MaskDataset(Dataset):
    """Base for datasets with mask annotation

    Define ``__getitem__`` method to load mask annotation paired with image.
    Pre-defined attributes are prefixed with a single underscore to distinguish
    them from those specific to a dataset. It is required to implement two
    methods: ``get_image()`` and ``get_mask()`` as well as ``acronym`` and
    ``_root_dir`` for each subclass.

    Attributes
    ----------
    output
    anno_dict

    Methods
    -------
    __getitem__

    Notes
    -----
    Required attributes in subclass
        - ``acronym``
        - ``_root_dir``
        - ``_output``
        - ``_grayscale`` (optional)
        - ``_grayscale_mode`` (optional)
        - ``_num_channels`` (optional)
        - ``get_image()``
        - ``get_mask()`` (optional)

    See Also
    --------
    Dataset : super class

    """
    @property
    def output(self) -> str:
        """Determine return(s) when called"""
        return self._output

    @output.setter
    def output(self, val):
        self._output = val

    @property
    def anno_dict(self) -> Dict[int, Any]:
        """Dictionary of pathes to annotation files"""
        raise NotImplementedError

    def __getitem__(self, ind: int) -> Dict[str, np.ndarray]:
        """Get image, mask, or both depending on ``output`` argument

        For MaskDataset, available output types are ['image', 'mask', 'both'].

        Parameters
        ----------
        ind : int
            Index to get path(s) from ``file_list`` attribute

        Attributes
        ----------
        self.output
        self.file_list
        self.anno_dict

        Other Parameters
        ----------------
        self._transforms
        self._num_calls
        self._grayscale
        self._grayscale_mode
        self._num_channels

        """
        # Randomize `ind` when `num_calls` set
        if self.num_calls is not None:
            if ind >= self.num_calls:
                raise IndexError('list index out of range')
            ind_max = len(self.file_list)
            if (self.output != 'image') and (self.anno_dict is not None):
                ind_max = len(self.anno_dict)
            ind = random.randrange(0, ind_max)
        # `output="image"`
        if self.output == 'image':
            p = self.file_list[ind]
            image = self.get_image(p)
            if self.grayscale:
                num_channels = self.num_channels
                if num_channels is None:
                    num_channels = len(p) if isinstance(p, list) else 3
                image = self.to_gray(
                    image,
                    grayscale_mode=self.grayscale_mode,
                    num_channels=num_channels
                )
            if self.transforms:
                image = self.transforms(image=image)['image']
            return {'image': image}
        # `output="mask"`
        elif self.output == 'mask':
            pm = self.anno_dict[ind]
            mask = self.get_mask(pm)
            if self.transforms is not None:
                mask = self.transforms(mask=mask)['mask']
                # # Filtering out empty masks
                # while mask.max() == 0:
                #     mask = self.transforms.augment_image(mask)
            return {'mask': mask}
        # both image and gt
        elif self.output == 'both':
            # 'image'
            p = self.file_list[ind]
            image = self.get_image(p)
            # 'mask'
            pm = self.anno_dict[ind]
            mask = self.get_mask(pm)
            if self.grayscale:
                num_channels = self.num_channels
                if num_channels is None:
                    num_channels = len(p) if isinstance(p, list) else 3
                image = self.to_gray(
                    image,
                    grayscale_mode=self.grayscale_mode,
                    num_channels=num_channels
                )
            # Make sure to apply the same augmentation both to image and mask
            if self.transforms is not None:
                augmented = self.transforms(image=image, mask=mask)
                image, mask = augmented['image'], augmented['mask']
            return {'image': image, 'mask': mask}
        else:
            raise NotImplementedError("Choose one ['image', 'mask', 'both']")

    def get_mask(self, key) -> np.ndarray:
        """Get a mask"""
        raise NotImplementedError


class IterDataset(Iterator):
    """Iterable
    """
    def __init__(self, dataset: Dataset):
        self.dataset = dataset
        self.ind = 0
        self.end = len(self.dataset)

    def __next__(self):
        if self.ind == self.end:
            raise StopIteration
        data = self.dataset[self.ind]
        self.ind += 1
        return data



import os.path
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional

import albumentations
import cv2
import numpy as np
import tifffile

from ..base import MaskDataset


class FRUNet(MaskDataset):
    """FRU-Net: Robust Segmentation of Small Extracellular Vesicles [1]_

    TEM images

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    normalize : bool, default: True
        Normalize each image by its maximum value and cast it to UINT8.

    Notes
    -----
    - Originally, dtype is UINT16
    - Max value is 20444, but contrast varies a lot. Normalization may be
      needed. Init param ``normalize`` is set to True by default for this
      reason.

    References
    ----------
    .. [1] E. Gómez-de-Mariscal, M. Maška, A. Kotrbová, V. Pospíchalová, P.
       Matula, and A. Muñoz-Barrutia, “Deep-Learning-Based Segmentation of Small
       Extracellular Vesicles in Transmission Electron Microscopy Images,”
       Scientific Reports, vol. 9, no. 1, Art. no. 1, Sep. 2019, doi:
       10.1038/s41598-019-49431-3.

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'FRUNet'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        # specific to this dataset
        normalize: bool = True,
        **kwargs
    ):
        self._root_dir = os.path.join(root_dir, 'code', 'data')
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        # specific to this dataset
        self.normalize = normalize

    def get_image(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        if self.normalize:
            v = tif.max()
            tif = 255 * (np.clip(tif, 0, v) / v)  # float64
            tif = tif.astype(np.uint8)
            tif = cv2.cvtColor(tif, cv2.COLOR_GRAY2RGB)
            return tif
        return tif

    def get_mask(self, p: Path) -> np.ndarray:
        mask = tifffile.imread(p)
        return mask.astype(np.int16)

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        file_list = sorted(root_dir.glob('dataset_*/*.tif'))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        anno_list = sorted(root_dir.glob('annotations_*/*.tif'))
        anno_dict = dict((k, v) for k, v in enumerate(anno_list))
        return anno_dict



from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import cv2
import numpy as np
import tifffile

from ..base import MaskDataset
from ..types import BundledPath
from ..utils import bundle_list, stack_channels, stack_channels_to_rgb


class BBBC008(MaskDataset):
    """Human HT29 colon-cancer cells [1]_

    F/B semantic segmentation

    The image set consists of 12 images. The samples were stained with Hoechst
    (channel 1), pH3 (channel 2), and phalloidin (channel 3). Hoechst labels
    DNA, which is present in the nucleus. Phalloidin labels actin, which is
    present in the cytoplasm. The last stain, pH3, indicates cells in mitosis;
    whereas this was important for Moffat et al.'s screen, it is irrelevant for
    segmentation and counting, so this channel is left out.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when `transforms` is set. Define the total length of the dataset.
        If it is set, it overrides ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    image_ch : {'DNA', 'actin'}, default: ('DNA', 'actin')
        Which channel(s) to load as image. Make sure to give it as a Sequence
        when choose a single channel.
    anno_ch : {'DNA', 'actin'}, default: ('DNA',)
        Which channel(s) to load as annotation. Make sure to give it as a
        Sequence when choose a single channel.

    Notes
    -----
    - Annotation F/B: BG=1, FG=0; very annoying...

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC008
    .. [2] Carpenter et al., Genome Biology, 2006

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC008'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        image_ch: Sequence[str] = ('DNA', 'actin',),
        anno_ch: Sequence[str] = ('DNA',),
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self.image_ch = image_ch
        self.anno_ch = anno_ch
        if not any([ch in ('DNA', 'actin') for ch in anno_ch]):
            raise ValueError("Set `anno_ch` in ('DNA', 'actin') in sequence")

    def get_image(self, p: Union[Path, BundledPath]) -> np.ndarray:
        if isinstance(p, Path):
            img = tifffile.imread(p)
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            # ch1 to red, ch3 to blue
            img = stack_channels_to_rgb(tifffile.imread, p, 0, 2, 1)
        return img

    def get_mask(self, p: Union[Path, BundledPath]) -> np.ndarray:
        if isinstance(p, Path):
            mask = tifffile.imread(p)
        else:
            mask = stack_channels(tifffile.imread, p)
        # dtype=bool originally and bool is not well handled by albumentations
        return 255 * (~mask).astype(np.uint8)

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        parent = 'human_ht29_colon_cancer_2_images'
        file_list = sorted(self.root_dir.glob(f'{parent}/*.tif'))
        if len(ch := self.image_ch) == 1:
            if ch[0] == 'DNA':
                file_list = file_list[::2]
            elif ch[0] == 'actin':
                file_list = file_list[1::2]
            else:
                raise ValueError("Set `anno_ch` in ('DNA', 'actin')")
            return file_list
        elif len(ch) == 2:
            file_blist = bundle_list(file_list, 2)
        else:
            raise ValueError("Set `anno_ch` in ('DNA', 'actin') or all")
        return file_blist

    @cached_property
    def anno_dict(self) -> Union[Dict[int, Path], Dict[int, BundledPath]]:
        parent = 'human_ht29_colon_cancer_2_foreground'
        anno_list = sorted(self.root_dir.glob(f'{parent}/*.tif'))
        if len(ch := self.anno_ch) == 1:
            if ch[0] == 'DNA':
                anno_list = anno_list[::2]
            elif ch[0] == 'actin':
                anno_list = anno_list[1::2]
            return dict((k, v) for k, v in enumerate(anno_list))
        elif len(ch) == 2:
            anno_blist = bundle_list(anno_list, 2)
        else:
            raise ValueError
        return dict((k, v) for k, v in enumerate(anno_blist))



import os.path
from functools import cached_property
from pathlib import Path
from typing import Dict, Optional

import albumentations
import cv2
import numpy as np
import tifffile

from ..base import MaskDataset


class S_BSST265(MaskDataset):
    """An annotated fluorescence image dataset for training nuclear segmentation
    methods [1]_

    Immuno Fluorescence (IF) images, designed for ML

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines
        augmentation in sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.

    Notes
    -----
    - All images have grayscale though some have 3 channels
    - rawimages: Raw nuclear images in TIFF format
    - groundtruth: Annotated masks in TIFF format
    - groundtruth_svgs: SVG-Files for each annotated masks and corresponding raw
      image in JPEG format
    - singlecell_groundtruth: Groundtruth for randomly selected nuclei of the
      testset (25 nuclei per testset class, a subset of all nuclei of the
      testset classes; human experts can compete with this low number of nuclei
      per subset by calculating Dice coefficients between their annotations and
      the groundtruth annotations)
    - visualized_groundtruth: Visualization of groundtruth masks in PNG format
    - visualized_singlecell_groundtruth: Visualization of groundtruth for
      randomly selected nuclei in PNG format
    - Find more info in README.txt inside the root directory

    References
    ----------
    .. [1] F. Kromp et al., “An annotated fluorescence image dataset for
       training nuclear segmentation methods,” Scientific Data, vol. 7, no. 1,
       Art. no. 1, Aug. 2020, doi: 10.1038/s41597-020-00608-w.

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'S_BSST265'

    def __init__(
        self,
        # Interface requirement
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        **kwargs
    ):
        # Interface and super-class arguments
        self._root_dir = os.path.join(root_dir, 'S-BSST265')
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls

    def get_image(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        if tif.shape[-1] != 3:
            tif = cv2.cvtColor(tif, cv2.COLOR_GRAY2RGB)
        return tif

    def get_mask(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        return tif.astype(np.int16)

    @cached_property
    def file_list(self) -> list:
        root_dir = self.root_dir
        parent = 'rawimages'
        file_list = sorted(
            root_dir.glob(f'{parent}/*.tif'), key=self._sort_key
        )
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        parent = 'groundtruth'
        anno_list = sorted(
            root_dir.glob(f'{parent}/*.tif'), key=self._sort_key
        )
        anno_dict = dict((k, v) for k, v in enumerate(anno_list))
        return anno_dict

    @staticmethod
    def _sort_key(p, zfill=2):
        split = p.stem.split('_')
        return '_'.join([s.zfill(zfill) for s in split])



"""(experimental) Easily load unknown datasets as Dataset or as MaskDataset

Common dataset is a dataset which has expected structures that
bioimageloader can easily access. Simply provide a path to root directory.

This module is experimental.

Currently it assumes two cases for MaskDataset,

>>> 1. case: only images
    case1/
    ├── image00.tif
    ├── image01.tif
    ├── image02.tif
    ├── image03.tif
    ├── image04.tif
    ├── image05.tif
    ├── image06.tif
    ├── image07.tif
    ├── image08.tif
    └── image09.tif

>>> 2. case: images in "images/" and labels in "labels/"
    case3/
    ├── images
    │   ├── 00.png
    │   ├── 01.png
    │   ├── 02.png
    │   ├── 03.png
    │   └── 04.png
    └── labels
        ├── 00.tif
        ├── 01.tif
        ├── 02.tif
        ├── 03.tif
        └── 04.tif

Examples
--------
Case 1:

>>> dataset = CommonDataset('./Data/case1')

Case 3:

>>> dataset = CommonMaskDataset('./Data/case3')

see also ``utils.get_maskdatasets_from_directory``

>>> datset = get_maskdataset_from_directory(
        './Data/case3',
        image_dir='images',
        labels='labels',
    )
"""

import os
from functools import cached_property
from pathlib import Path
from typing import Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile

from .base import Dataset, MaskDataset
from .types import KNOWN_IMAGE_EXT, PIL_IMAGE_EXT, TIFFFILE_IMAGE_EXT
from .utils import imread_asarray


class CommonDataset(Dataset):
    """Load a dataset thas has a common structure

    Parameters
    ----------
    root_dir
    output : optional
    transforms : optional
    num_calls : optional
    grayscale : optional
    grayscale_mode : optional
    num_channels : optional

    Attributes
    ----------
    image_dir
    file_list

    Methods
    -------
    get_image
    _setattr_ifvalue
    _filter_known_ext

    See Also
    --------
    Dataset : super class
    bioimageloader.utils.get_dataset_from_directory : util

    """
    count = 0
    acronym = 'dataset'

    def __init__(
        self,
        root_dir,
        *,
        output: Optional[str] = None,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: Optional[bool] = None,
        grayscale_mode: Optional[Union[str, Sequence[float]]] = None,
        **kwargs
    ):
        self.acronym = f'dataset_{CommonMaskDataset.count}'
        self._root_dir = root_dir
        # keywords
        self._setattr_ifvalue('_output', output)
        self._setattr_ifvalue('_transforms', transforms)
        self._setattr_ifvalue('_num_calls', num_calls)
        self._setattr_ifvalue('_grayscale', grayscale)
        self._setattr_ifvalue('_grayscale_mode', grayscale_mode)
        # count # of instances
        CommonDataset.count += 1

    def _setattr_ifvalue(self, attr, value=None):
        """Set attribute if value is not None"""
        if value is not None:
            setattr(self, attr, value)

    @staticmethod
    def _filter_known_ext(p: Path):
        """Filter extensions supported by PIL and tifffile"""
        return p.suffix.lower() in KNOWN_IMAGE_EXT

    @cached_property
    def file_list(self):
        return sorted(filter(self._filter_known_ext, self.root_dir.iterdir()))

    def get_image(self, p: Path) -> np.ndarray:
        if (suffix := p.suffix.lower()) in TIFFFILE_IMAGE_EXT:
            img = tifffile.imread(p)
        elif suffix in PIL_IMAGE_EXT:
            img = imread_asarray(p)
        return img


class CommonMaskDataset(CommonDataset, MaskDataset):
    """Load a dataset thas has a common structure with its mask annotation

    Call this from ``bioimageloader.utils.get_maskdataset_from_directory()``

    Parameters
    ----------
    root_dir
    output : optional
    transforms : optional
    num_calls : optional
    grayscale : optional
    grayscale_mode : optional
    num_channels : optional

    Attributes
    ----------
    image_dir
    mask_dir
    file_list
    anno_dict

    Methods
    -------
    get_image
    get_mask

    See Also
    --------
    MaskDataset : super class
    CommonDataset : super class
    bioimageloader.utils.get_maskdataset_from_directory : util

    """
    count = 0
    acronym = 'maskdataset'

    def __init__(
        self,
        root_dir,
        *,
        output: Optional[str] = None,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: Optional[bool] = None,
        grayscale_mode: Optional[Union[str, Sequence[float]]] = None,
        **kwargs
    ):
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            grayscale=grayscale,
            grayscale_mode=grayscale_mode,
            **kwargs
        )
        self.acronym = f'maskdataset_{CommonMaskDataset.count}'
        # count # of instances
        CommonMaskDataset.count += 1

    @property
    def image_dir(self) -> Optional[Path]:
        if hasattr(self, '_image_dir'):
            if (_image_dir := self.root_dir / self._image_dir).is_dir():
                return _image_dir
            else:
                raise NotADirectoryError
        if (n := 'images') in os.listdir(self.root_dir):
            if (_image_dir := self.root_dir / n).is_dir():
                return _image_dir
        return None

    @image_dir.setter
    def image_dir(self, val):
        self._image_dir = val

    @property
    def mask_dir(self) -> Optional[Path]:
        if hasattr(self, '_mask_dir'):
            if (_mask_dir := self.root_dir / self._mask_dir).is_dir():
                return _mask_dir
            else:
                raise NotADirectoryError
            return self.root_dir / self._mask_dir
        if (n := 'labels') in os.listdir(self.root_dir):
            if (_mask_dir := self.root_dir / n).is_dir():
                return _mask_dir
        return None

    @mask_dir.setter
    def mask_dir(self, val):
        self._mask_dir = val

    @cached_property
    def file_list(self):
        image_dir = self.image_dir if self.image_dir else self.root_dir
        return sorted(filter(self._filter_known_ext, image_dir.iterdir()))

    @cached_property
    def anno_dict(self):
        mask_dir = self.mask_dir if self.mask_dir else self.root_dir
        return sorted(filter(self._filter_known_ext, mask_dir.iterdir()))

    def get_mask(self, p: Path) -> np.ndarray:
        if (suffix := p.suffix.lower()) in TIFFFILE_IMAGE_EXT:
            img = tifffile.imread(p)
        elif suffix in PIL_IMAGE_EXT:
            img = imread_asarray(p)
        return img



from pathlib import Path
from typing import List, TypeVar, NewType, Any

T = TypeVar('T')

# Keep all ext in lower case
PIL_IMAGE_EXT = [
    '.png',
    '.jpg',
    '.jpeg',
]
TIFFFILE_IMAGE_EXT = [
    '.tif',
    '.tiff',
]
KNOWN_IMAGE_EXT = PIL_IMAGE_EXT + TIFFFILE_IMAGE_EXT
Bundled = List[T]
BundledPath = Bundled[Path]



from pathlib import Path
from typing import List, Optional, Sequence, Union

import albumentations
import numpy as np
from PIL import Image
from scipy.ndimage import binary_fill_holes

from ..utils import stack_channels
from bioimageloader.collections import BBBC018


class BBBC018NUnet(BBBC018):
    """Human HT29 colon-cancer cells (diverse phenotypes) (Adopted for NU-Net)

    The image set consists of 56 fields of view (4 from each of 14 samples).
    Because there are three channels, there are 168 image files. (The samples
    were stained with Hoechst 33342, pH3, and phalloidin. Hoechst 33342 is a DNA
    stain that labels the nucleus. Phospho-histone H3 indicates mitosis.
    Phalloidin labels actin, which is present in the cytoplasm.) The samples are
    the top-scoring sample from each of Jones et al.'s classifiers, as listed in
    the file SamplesScores.zip in their supplement. The files are in DIB format,
    as produced by the Cellomics ArrayScan instrument at the Whitehead–MIT
    Bioimaging Center. We recommend using Bio-Formats to read the DIB files.
    Each image is 512 x 512 pixels.

    The filenames are of the form wellidx-channel.DIB, where wellidx is the
    five-digit well index (from Jones et al.'s supplement) and channel is either
    DNA, actin, or pH3, depending on the channel.

    Notes
    -----
    - BBBC018_v1_images/10779 annotation is missing. len(anno_dict) =
      len(file_list) - 1; ind={26}
        [PosixPath('images/bbbc/018/BBBC018_v1_images/10779-DNA.DIB'),
         PosixPath('images/bbbc/018/BBBC018_v1_images/10779-actin.DIB'),
         PosixPath('images/bbbc/018/BBBC018_v1_images/10779-pH3.DIB')]
    - Every DIB has 3 channels (Order = (DNA,actin,pH3)). The second one is the
      object.
    - DNA -> Nuceli
    - Actin -> Cell
    - Annotation is outline one, but every anno is closed so binary_fill_holes
      works fine
    - For some reason annotation is y inverted

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC018
    """

    # Dataset's acronym
    acronym = 'BBBC018'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        anno_ch: Sequence[str] = ('DNA',),
        drop_missing_pairs: bool = True,
        fill_holes: bool = True,
        **kwargs
    ):
        """
        Parameters
        ----------
        root_dir : str
            Path to root directory
        output : {'image', 'mask', 'both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        grayscale : bool (default: False)
            Convert images to grayscale
        grayscale_mode : {'cv2', 'equal', Sequence[float]} (default: 'equal')
            How to convert to grayscale. If set to 'cv2', it follows opencv
            implementation. Else if set to 'equal', it sums up values along
            channel axis, then divides it by the number of expected channels.
        anno_ch : {'DNA', 'actin'} (default: ('DNA',))
            Which channel(s) to load as annotation. Make sure to give it as a
            Sequence when choose a single channel.
        drop_missing_pairs : bool (default: True)
            Valid only if `output='both'`. It will drop images that do not have
            mask pairs.
        fill_holes : bool (default: True)
            Fill outline annotation using `scipy.ndimage.binary_fill_holes()`

        See Also
        --------
        BBBC018 : Super class
        MaskDataset : Super class
        DatasetInterface : Interface
        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            grayscale=grayscale,
            grayscale_mode=grayscale_mode,
            anno_ch=anno_ch,
            drop_missing_pairs=drop_missing_pairs,
            **kwargs
        )
        self.fill_holes = fill_holes

    def get_mask(self, p: Union[Path, List[Path]]) -> np.ndarray:
        if isinstance(p, Path):
            mask = np.asarray(Image.open(p))
            if self.fill_holes:
                mask = binary_fill_holes(mask)
        else:
            mask = stack_channels(Image.open, p)
            if self.fill_holes:
                for c in range(mask.shape[-1]):
                    mask[..., c] = binary_fill_holes(mask[..., c])
        # For some reason mask is -y
        return np.ascontiguousarray(mask[::-1, ...]).astype(np.float32)



from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np

from bioimageloader.base import Dataset, MaskDataset
from bioimageloader.types import BundledPath


class DatasetTemplate(Dataset):
    """Template for Dataset

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.

    See Also
    --------
    MaskDataset : Super class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = ''

    def __init__(
        self,
        root_dir: str,
        *,  # only keyword param
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,  # optional
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',  # optional
        # specific to this dataset
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale   # optional
        self._grayscale_mode = grayscale_mode  # optional
        # specific to this one here

    def get_image(self, p: Union[Path, List[BundledPath]]) -> np.ndarray:
        ...

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        # Important to decorate with `cached_property` in general
        ...


class MaskTemplate(MaskDataset):
    """Template for MaskDataset

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.

    See Also
    --------
    MaskDataset : Super class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = ''

    def __init__(
        self,
        root_dir: str,
        *,  # only keyword param
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,  # optional
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',  # optional
        # specific to this dataset
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale   # optional
        self._grayscale_mode = grayscale_mode  # optional
        # specific to this one here

    def get_image(self, p: Union[Path, List[BundledPath]]) -> np.ndarray:
        ...

    def get_mask(self, p: Union[Path, List[BundledPath]]) -> np.ndarray:
        ...

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        # Important to decorate with `cached_property` in general
        ...

    @cached_property
    def anno_dict(self) -> Union[Dict[int, Path], Dict[int, BundledPath]]:
        # Important to decorate with `cached_property` in general
        ...



from pathlib import Path
from typing import List, Optional, Sequence, Union

import albumentations
import numpy as np
import scipy.ndimage as ndi
import tifffile

from bioimageloader.collections import BBBC007
from bioimageloader.utils import stack_channels


class BBBC007NUnet(BBBC007):
    """Drosophila Kc167 cells (Adopted for NU-Net)

    NU-Net needs outline to be filled and only need DNA annotation.

    Outline annotation

    Images were acquired using a motorized Zeiss Axioplan 2 and a Axiocam MRm
    camera, and are provided courtesy of the laboratory of David Sabatini at the
    Whitehead Institute for Biomedical Research. Each image is roughly 512 x 512
    pixels, with cells roughly 25 pixels in dimeter, and 80 cells per image on
    average. The two channels (DNA and actin) of each image are stored in
    separate gray-scale 8-bit TIFF files.

    Notes
    -----
    - [4, 5, 11, 14, 15] have 3 channels but they are just all gray scale
        images. Extra work is required in get_image().

    .. [1] Jones et al., in the Proceedings of the ICCV Workshop on Computer
       Vision for Biomedical Image Applications (CVBIA), 2005.
    .. [2] [BBBC007](https://bbbc.broadinstitute.org/BBBC007)
    """
    # Dataset's acronym
    acronym = 'BBBC007'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        # specific to this dataset
        image_ch: Sequence[str] = ('DNA', 'actin',),
        anno_ch: Sequence[str] = ('DNA',),
        fill_holes: bool = True,
        **kwargs
    ):
        """
        Parameters
        ----------
        root_dir : str
            Path to root directory
        output : {'image', 'mask', 'both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        image_ch : {'DNA', 'actin'} (default: ('DNA', 'actin'))
            Which channel(s) to load as image. Make sure to give it as a
            Sequence when choose a single channel.
        anno_ch : {'DNA', 'actin'} (default: ('DNA',))
            Which channel(s) to load as annotation. Make sure to give it as a
            Sequence when choose a single channel.
        fill_holes : bool (default: True)
            Fill outline annotation using `scipy.ndimage.binary_fill_holes()`

        See Also
        --------
        BBBC007 : Super class
        MaskDataset : Super class
        DatasetInterface : Interface
        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            image_ch=image_ch,
            anno_ch=anno_ch,
            **kwargs
        )
        self.fill_holes = fill_holes

    def get_mask(self, p: Union[Path, List[Path]]) -> np.ndarray:
        if isinstance(p, Path):
            mask = tifffile.imread(p)
        else:
            mask = stack_channels(tifffile.imread, p)
        # dtype=bool originally and bool is not well handled by albumentations
        if self.fill_holes:
            mask = ndi.binary_fill_holes(mask)
        return mask.astype(np.float32)



"""Root module of ``bioimageloader``

Expose core classes and functions
"""

from .batch import BatchDataloader, ConcatDataset
from .config import Config

__all__ = [
    'BatchDataloader',
    'ConcatDataset',
    'Config',
]



import os.path
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile
from PIL import Image

from ..base import MaskDataset


class DigitalPathology(MaskDataset):
    """Deep learning for digital pathology image analysis: A comprehensive
    tutorial with selected use cases [1]_

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.

    Notes
    -----
    - Annotation is partial
    - Boolean mask to UINT8 mask (0, 255)

    References
    ----------
    .. [1] A. Janowczyk and A. Madabhushi, “Deep learning for digital pathology
       image analysis: A comprehensive tutorial with selected use cases,” J
       Pathol Inform, vol. 7, Jul. 2016, doi: 10.4103/2153-3539.186902.

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'DigitPath'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        **kwargs
    ):
        self._root_dir = os.path.join(root_dir, 'nuclei')
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode

    def get_image(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        return tif

    def get_mask(self, p: Path) -> np.ndarray:
        mask = np.asarray(Image.open(p))
        return 255 * mask.astype(np.uint8)

    def __len__(self):
        if self.num_calls:
            return self.num_calls
        return len(self.file_list)

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = 'original'
        file_list = sorted(root_dir.glob(f'{parent}/*.tif'))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        parent = 'mask'
        anno_list = sorted(root_dir.glob(f'{parent}/*.png'))
        return dict((k, v) for k, v in enumerate(anno_list))



"""Batch loader module
"""

import bisect
import concurrent.futures
import random
import warnings
from math import ceil
from typing import Dict, Iterator, List, Optional

import numpy as np

from .base import Dataset


class ConcatDataset:
    """Concatenate Datasets

    Todo
    ----
    - Typing datasets with covariant class
        Lose param hints because of Generic type
    - Intermediate class linking DatasetInterface and [MaskDataset, BBoxDataset,
      BBoxDataset, ...]

    References
    ----------
    .. [1] https://pytorch.org/docs/stable/_modules/torch/utils/data/dataset.html#ConcatDataset
    """
    def __init__(self, datasets: List[Dataset]):
        self.datasets = datasets
        self.acronym = [dset.acronym for dset in self.datasets]
        self.sizes = [len(dset) for dset in self.datasets]
        self.cumulative_sizes = np.cumsum(self.sizes)
        # Check and Warn
        if any([s == 0 for s in self.sizes]):
            i = self.sizes.index(0)
            warnings.warn(f"ind={i} {self.datasets[i].acronym} is empty")
        if len(set(outputs := [dset.output for dset in self.datasets])) != 1:
            warnings.warn(f"output types do not match {outputs}")

    def __len__(self):
        return self.cumulative_sizes[-1]

    def __getitem__(self, ind):
        ind_dataset = bisect.bisect_right(self.cumulative_sizes, ind)
        ind_sample = ind if ind_dataset == 0 else ind - self.cumulative_sizes[ind_dataset - 1]
        return self.datasets[ind_dataset][ind_sample]


class BatchDataloader:
    """Batch loader with multi-processing
    """
    def __init__(
        self,
        dataset: Dataset,
        batch_size: int = 16,
        shuffle: bool = False,
        drop_last: bool = False,
        num_workers: Optional[int] = None,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.num_workers = num_workers
        self.executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.num_workers)

        if shuffle:
            idx = list(range(len(self.dataset)))
            random.shuffle(idx)
            dataset.file_list = [dataset.file_list[ind] for ind in idx]
            if hasattr(dataset, 'anno_dict'):
                dataset.anno_dict = dict((i, dataset.anno_dict[ind])
                                         for i, ind in enumerate(idx))

    def __len__(self):
        if self.drop_last:
            return len(self.dataset) // self.batch_size
        return ceil(len(self.dataset) / self.batch_size)

    @property
    def _last_size(self):
        if self.drop_last:
            return self.batch_size
        remainder = len(self.dataset) - self.batch_size * (len(self) - 1)
        return remainder

    def __iter__(self):
        return IterBatchDataloader(self)

    def __del__(self):
        # Shutdown mp
        self.executor.shutdown(wait=True)


class IterBatchDataloader(Iterator):
    """Iterate BatchDataloader
    """
    def __init__(self, dataloader: BatchDataloader):
        self.dataloader = dataloader
        self.dataset = dataloader.dataset
        self.batch_size = self.dataloader.batch_size
        self.drop_last = self.dataloader.drop_last
        self.num_workers = self.dataloader.num_workers
        self.executor = self.dataloader.executor
        self._last_size = self.dataloader._last_size

        self.batch_ind = 0
        self.batch_end = len(dataloader)

    def __next__(self) -> Dict[str, np.ndarray]:
        if self.batch_ind == self.batch_end:
            raise StopIteration
        if self.batch_ind == self.batch_end - 1 and not self.drop_last:
            batch_size = self._last_size
        else:
            batch_size = self.batch_size
        ind_start = self.batch_size * self.batch_ind
        jobs = [self.executor.submit(_mp_getitem, self.dataset, ind)
                for ind in range(ind_start, ind_start + batch_size)]
        concurrent.futures.wait(jobs)
        self.batch_ind += 1
        # return [j.result() for j in jobs]
        # bundle with the same keys
        d0 = jobs[0].result()
        keys = d0.keys()
        ret = {}
        for k in keys:
            ret[k] = np.stack([d0[k]] + list(map(lambda j: j.result()[k],
                                                 jobs[1:]))
                              )
        return ret


def _mp_getitem(dataset, ind):
    return dataset[ind]



"""StarDist data[1]_ is a subset of DSB2018

.. [1] https://github.com/stardist/stardist/releases/download/0.1.0/dsb2018.zip
"""

import os.path
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union, Any

import albumentations
import cv2
import numpy as np
import tifffile

from ..base import MaskDataset
from ..types import BundledPath
from ..utils import imread_asarray, rle_decoding_inseg, read_csv, ordered_unique


class StarDist(MaskDataset):
    """Training set for StarDist [1]_, [2]_

    StarDist data is a subset of Data Science Bowl 2018 [3]_

    Find the nuclei in divergent images to advance medical discovery

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    training : bool, default: True
        Load training set if True, else load testing one

    References
    ----------
    .. [1] https://arxiv.org/abs/1806.03535
    .. [2] https://github.com/stardist/stardist/releases/download/0.1.0/dsb2018.zip
    .. [3] https://www.kaggle.com/c/data-science-bowl-2018/

    See Also
    --------
    DSB2018
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = 'StarDist'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # specific to this dataset
        training: bool = True,
        **kwargs
    ):
        self._root_dir = os.path.join(root_dir, 'dsb2018')
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this one here
        self.training = training

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        # img = Image.open(p)
        # img = img.convert(mode='RGB')
        # return np.asarray(img)

    def get_mask(self, p: Path) -> np.ndarray:
        return tifffile.imread(p)
        # if self.training and not isinstance(anno, dict):
        #     # anno: BundlePath
        #     p = anno[0]
        #     val = 1
        #     m0 = imread_asarray(p) > 0
        #     mask = np.zeros_like(m0, dtype=np.uint8)  # uint8 is enough
        #     mask[m0] = val
        #     for p in anno[1:]:
        #         val += 1
        #         m = imread_asarray(p) > 0
        #         # Does not allow overlapping!
        #         mask[m] = val
        #     return mask
        # # anno: dict
        # run_lengths = anno['EncodedPixels']
        # h, w = anno['Height'], anno['Width']
        # mask = rle_decoding_inseg((h, w), run_lengths)
        # return mask

    # @cached_property
    # def ids(self) -> List[str]:
    #     if self.training:
    #         self.root_dir / 'train' /
    #     ids = [line[0] for line in lines]
    #     ids = ordered_unique(ids)
    #     return ids

    @cached_property
    def file_list(self) -> List[Path]:
        # Call MaskDataset.root_dir
        parent = 'train' if self.training else 'test'
        return sorted((self.root_dir / parent / 'images').glob('*.tif'))

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        return dict((i, p.parent.with_name('masks') / p.name)
                    for i, p in enumerate(self.file_list))
        # if self.training:
        #     anno_dict = {}
        #     for i, p in enumerate(self.file_list):
        #         anno_dict[i] = list(p.parents[1].glob('masks/*.png'))
        #     return anno_dict
        # else:
        #     anno_rle = {}
        #     _, lines = read_csv(self.root_dir / 'stage1_solution.csv')
        #     # header: ImageId,EncodedPixels,Height,Width,Usage
        #     # iter_rle = map(lambda line: [int(s) for s in line[1].split(' ')],
        #     #                lines)
        #     offset = 0
        #     for i, idx in enumerate(self.ids):
        #         solution: dict = {'EncodedPixels': []}
        #         for line in lines[offset:]:
        #             if idx == line[0]:
        #                 if 'Height' not in solution:
        #                     solution['Height'] = int(line[2])
        #                     solution['Width'] = int(line[3])
        #                 solution['EncodedPixels'].append(
        #                     [int(s) for s in line[1].split(' ')]
        #                 )
        #                 offset += 1
        #             else:
        #                 break
        #         anno_rle[i] = solution
        #     return anno_rle



from typing import List, Optional, Union, Dict

import albumentations
import yaml

from .base import Dataset
from .collections import *


class Config(dict):
    """Construct config from a yaml file

    Parameters
    ----------
    filename : str
        Path to config file
    """
    def __init__(self, filename):
        self.filename = filename
        cfg = self._read_yaml()
        for k, v in cfg.items():
            self[k] = v

    def _read_yaml(self) -> dict:
        with open(self.filename, 'r') as f:
            cfg = yaml.safe_load(f)
        return cfg

    @staticmethod
    def from_dict(d: dict):
        return _Config(d)

    def load_datasets(
        self,
        transforms: Optional[Union[albumentations.Compose, Dict[str, albumentations.Compose]]] = None,
    ) -> List[Dataset]:
        """Load multiple datasets from a yaml file

        Parameters
        ----------
        config : configuration object
            Config instance that contains acronyms and arguments to initialize
            each dataset
        transforms : albumentations.Compose or dictionary, optional
            Either apply a single composed transformations for every datasets or
            pass a dictionary that defines transformations for each dataset with
            keys being the acronyms of datsets.

        """
        datasets: List[Dataset] = []
        for dataset, kwargs in self.items():
            if isinstance(transforms, dict):
                exec(f'datasets.append({dataset}(transforms=transforms[dataset], **kwargs))')
            else:
                exec(f'datasets.append({dataset}(transforms=transforms, **kwargs))')
        return datasets


class _Config(Config):
    """Config.from_dict()"""
    def __init__(self, d: dict):
        for k, v in d.items():
            self[k] = v



from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union, overload

import albumentations
import cv2
import numpy as np
import tifffile

from ..base import MaskDataset
from ..types import BundledPath
from ..utils import bundle_list, stack_channels, stack_channels_to_rgb


class BBBC007(MaskDataset):
    """Drosophila Kc167 cells

    Outline annotation

    Images were acquired using a motorized Zeiss Axioplan 2 and a Axiocam MRm
    camera, and are provided courtesy of the laboratory of David Sabatini at the
    Whitehead Institute for Biomedical Research. Each image is roughly 512 x 512
    pixels, with cells roughly 25 pixels in dimeter, and 80 cells per image on
    average. The two channels (DNA and actin) of each image are stored in
    separate gray-scale 8-bit TIFF files.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'equal', 'cv2', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    image_ch : {'DNA', 'actin'}, default: ('DNA', 'actin')
        Which channel(s) to load as image. Make sure to give it as a Sequence
        when choose a single channel. Name matches to `anno_ch`.
    anno_ch : {'DNA', 'actin'}, default: ('DNA',)
        Which channel(s) to load as annotation. Make sure to give it as a
        Sequence when choose a single channel.

    Notes
    -----
    - [4, 5, 11, 14, 15] have 3 channels but they are just all gray scale
      images. Extra work is required in get_image().

    References
    ----------
    .. [1] Jones et al., in the Proceedings of the ICCV Workshop on Computer
       Vision for Biomedical Image Applications (CVBIA), 2005.
    .. [2] https://bbbc.broadinstitute.org/BBBC007

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC007'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        image_ch: Sequence[str] = ('DNA', 'actin',),
        anno_ch: Sequence[str] = ('DNA',),
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        self.image_ch = image_ch
        self.anno_ch = anno_ch
        if not any([ch in ('DNA', 'actin') for ch in image_ch]):
            raise ValueError("Set `anno_ch` in ('nuclei', 'cells') in sequence")
        if not any([ch in ('DNA', 'actin') for ch in anno_ch]):
            raise ValueError("Set `anno_ch` in ('nuclei', 'cells') in sequence")

    @staticmethod
    def _imread_handler(p: Path) -> np.ndarray:
        """Handle irregular images by wrapping tifffile.imread

        Normally two images in a pair have gray scale and only have one channel.
        This means that each image array has shape of (height, width). But there
        are some outliers.

        For example a pair of images below has 3 channels with all having the
        same value (height, width, 3).
        ['BBBC007_v1_images/f113/AS_09125_040701150004_A02f00d0.tif',
         'BBBC007_v1_images/f113/AS_09125_040701150004_A02f00d1.tif']

        6 pairs out of 16 have this issue and this wrapper resolves it.
        """
        img = tifffile.imread(p)
        if img.shape[-1] == 3:
            return img[..., 0]
        return img

    def get_image(self, p: Union[Path, BundledPath]) -> np.ndarray:
        if isinstance(p, Path):
            img = self._imread_handler(p)
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            img = stack_channels_to_rgb(self._imread_handler, p)
        return img

    def get_mask(self, p: Union[Path, BundledPath]) -> np.ndarray:
        if isinstance(p, Path):
            mask = tifffile.imread(p)
        else:
            mask = stack_channels(tifffile.imread, p)
        # dtype=bool originally and bool is not well handled by albumentations
        return 255 * mask.astype(np.uint8)

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        file_list: Union[List[Path], List[List[Path]]]
        root_dir = self.root_dir
        parent = 'BBBC007_v1_images'
        _file_list = sorted(root_dir.glob(f'{parent}/*/*.tif'))
        if len(ch := self.image_ch) == 1:
            if ch[0] == 'DNA':
                file_list = _file_list[::2]
            elif ch[0] == 'actin':
                file_list = _file_list[1::2]
            else:
                raise ValueError("Set `anno_ch` in ('DNA', 'actin')")
        elif len(ch) == 2:
            file_list = bundle_list(_file_list, 2)
        else:
            raise ValueError("Set `anno_ch` in ('DNA', 'actin') or all")
        return file_list

    @cached_property
    def anno_dict(self) -> Union[Dict[int, Path], Dict[int, BundledPath]]:
        root_dir = self.root_dir
        parent = 'BBBC007_v1_outlines'
        _anno_list = sorted(root_dir.glob(f'{parent}/*/*.tif'))
        if len(ch := self.anno_ch) == 1:
            if ch[0] == 'DNA':
                anno_list = _anno_list[::2]
            elif ch[0] == 'actin':
                anno_list = _anno_list[1::2]
            return dict((k, v) for k, v in enumerate(anno_list))
        elif len(ch) == 2:
            anno_blist = bundle_list(_anno_list, 2)
        else:
            raise ValueError("Set `anno_ch` in ('DNA', 'actin') or all")
        return dict((k, v) for k, v in enumerate(anno_blist))



from functools import cached_property
from pathlib import Path
from typing import List, Optional, Sequence, Union

import albumentations
import cv2
import numpy as np
from PIL import Image

from ..base import Dataset
from ..types import BundledPath
from ..utils import bundle_list, imread_asarray, stack_channels_to_rgb


class BBBC014(Dataset):
    """Human U2OS cells cytoplasm–nucleus translocation

    This 96-well plate has images of cytoplasm to nucleus translocation of the
    transcription factor NFκB in MCF7 (human breast adenocarcinoma cell line)
    and A549 (human alveolar basal epithelial) cells in response to TNFα
    concentration.

    Images are at 10x objective magnification. The plate was acquired at Vitra
    Bioscience on the CellCard reader. For each well there is one field with two
    images: a nuclear counterstain (DAPI) image and a signal stain (FITC) image.
    Image size is 1360 x 1024 pixels. Images are in 8-bit BMP format.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'equal', 'cv2', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    image_ch : {'DAPI', 'FITC'}, default: ('DAPI', 'FITC')
        Which channel(s) to load as image. Make sure to give it as a Sequence
        when choose a single channel.

    Notes
    -----
    - Second channel is usually very clear with a few artifacts
    - Biological annotation
    - CellProfiler's LoadText module format annotation also available (not
      implemented)
    - Zoom in?

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC014

    See Also
    --------
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC014'

    def __init__(
        self,
        root_dir: str,
        *,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        image_ch: Sequence[str] = ('DAPI', 'FITC'),
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self.image_ch = image_ch
        if not any([ch in ('DAPI', 'FITC') for ch in image_ch]):
            raise ValueError("Set `image_ch` in ('DAPI', 'FITC') in sequence")

    def get_image(self, p: Union[Path, BundledPath]) -> np.ndarray:
        if isinstance(p, Path):
            img = imread_asarray(p)
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            img = stack_channels_to_rgb(Image.open, p, 1, 2, 0)
        return img

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        root_dir = self.root_dir
        parent = 'BBBC014_v1_images'
        file_list = sorted(root_dir.glob(f'{parent}/*.Bmp'), key=self._sort_key)
        if len(ch := self.image_ch) == 1:
            if ch[0] == 'DAPI':
                return file_list[::2]
            elif ch[0] == 'FITC':
                return file_list[1::2]
        return bundle_list(file_list, 2)

    @staticmethod
    def _sort_key(p: Path):
        channel, ind, t, subind, _ = p.stem.split('-')
        return '-'.join([ind, t, subind, channel])



import re
from functools import cached_property
from pathlib import Path
from typing import List, Optional, Sequence, Union

import albumentations
import cv2
import numpy as np
import tifffile

from ..base import Dataset
from ..types import BundledPath
from ..utils import bundle_list, stack_channels_to_rgb


class BBBC021(Dataset):
    """Human MCF7 cells – compound-profiling experiment [1]_

    The images are of MCF-7 breast cancer cells treated for 24 h with a
    collection of 113 small molecules at eight concentrations. The cells were
    fixed, labeled for DNA, F-actin, and Β-tubulin, and imaged by fluorescent
    microscopy as described [Caie et al. Molecular Cancer Therapeutics, 2010].

    There are 39,600 image files (13,200 fields of view imaged in three
    channels) in TIFF format. We provide the images in 55 ZIP archives, one for
    each microtiter plate. The archives are ~750 MB each.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'equal', 'cv2', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    image_ch : {'DNA', 'actin'}, default: ('DNA', 'actin', 'tublin')
        Which channel(s) to load as image. Make sure to give it as a Sequence
        when choose a single channel.

    Notes
    -----
    - HUGE dataset
    - 3 channels
        - w1 (DNA) -> Blue
        - w2 (actin?) -> Green
        - w4 (tublin??)-> Red
    - UINT16

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC021

    See Also
    --------
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC021'

    def __init__(
        self,
        root_dir: str,
        *,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # # specific to this dataset
        image_ch: Sequence[str] = ('DNA', 'actin', 'tublin'),
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        self.image_ch = image_ch
        if not any([ch in ('DNA', 'actin', 'tublin') for ch in image_ch]):
            raise ValueError("Set `anno_ch` in ('DNA', 'actin', 'tublin') in sequence")

    def get_image(self, p: Union[Path, List[Path]]) -> np.ndarray:
        # 3 channels; DAPI(w1), Tubulin(w2), Actin(w4)
        if isinstance(p, Path):
            img = tifffile.imread(p)
            img = (img / 2**8).astype(np.uint8)
            return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        if len(self.image_ch) == 2:
            def _map_ch_to_ind(p: Path):
                if 'w1' in p.stem:
                    # w1 'DNA' blue
                    return 2
                elif 'w2' in p.stem:
                    # w2 'actin' green
                    return 1
                # w4 'tublin' red
                return 0
            order = map(_map_ch_to_ind, p)
            img = stack_channels_to_rgb(tifffile.imread, p, *order)
            img = (img / 2**8).astype(np.uint8)
            return img
        img = stack_channels_to_rgb(tifffile.imread, p, 2, 1, 0)
        img = (img / 2**8).astype(np.uint8)
        return img

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        root_dir = self.root_dir
        _file_list = sorted(root_dir.glob('Week*/*.tif'))
        if len(ch := self.image_ch) == 1:
            if ch[0] == 'DNA':
                return _file_list[::3]
            elif ch[0] == 'actin':
                return _file_list[1::3]
            elif ch[0] == 'tublin':
                return _file_list[2::3]
            else:
                raise ValueError
        elif len(ch) == 2:
            map_to_pat = {
                'DNA': 'w1',     # blue
                'actin': 'w2',   # green
                'tublin': 'w4',  # red
            }
            # regex pattern
            pat = '(' + '|'.join(map_to_pat[c] for c in ch) + ')'
            return bundle_list(
                list(filter(lambda p: re.search(pat, p.stem), _file_list)), 2
            )
        return bundle_list(_file_list, 3)



from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile
from PIL import Image

from ..base import MaskDataset
from ..types import BundledPath
from ..utils import bundle_list, stack_channels_to_rgb


class BBBC006(MaskDataset):
    """Human U2OS cells (out of focus)

    Images were acquired from one 384-well microplate containing U2OS cells
    stained with Hoechst 33342 markers (to label nuclei) were imaged with an
    exposure of 15 and 1000 ms for Hoechst and phalloidin respectively, at 20x
    magnification, 2x binning, and 2 sites per well. For each site, the optimal
    focus was found using laser auto-focusing to find the well bottom. The
    automated microscope was then programmed to collect a z-stack of 32 image
    sets (z = 16 at the optimal focal plane, 15 images above the focal plane, 16
    below) with 2 μm between slices. Each image is 696 x 520 pixels in 16-bit
    TIF format, LZW compression. Each image filename includes either 'w1' to
    denote Hoechst images or 'w2' to denote phalloidin images.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'equal', 'cv2', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    uint8 : bool, default: True
        Whether to convert images to UINT8. It will divide image by 2**12 and
        cast it to UINT8. If set False, no process will be applied. Read more
        about rationales in Notes section.
    z_ind : int, default: 16
        Select one z stack. Default is 16, because 16 is the most in-focus.

    Notes
    -----
    - z-stack, z=16 is in-focus ones, sites (s1, s2)
    - Instance segmented
    - 384 wells, 2 sites per well; 384 * 2 = 768 images
    - 2 channels, w1=Hoechst, w2=phalloidin
    - Two channels usually overlap and when overlapped, it's hard to distinguish
      two channels anymore.
    - Saved in UINT16, but UINT12 practically. Max value caps at 4095.

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC006

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC006'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        uint8: bool = True,
        z_ind: int = 16,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self.uint8 = uint8
        self.z_ind = z_ind

    def get_image(self, p: BundledPath) -> np.ndarray:
        # 2 channels
        img = stack_channels_to_rgb(tifffile.imread, p, 2, 0, 1)
        # UINT12
        if self.uint8:
            img = (img / 2**4).astype(np.uint8)
        return img

    def get_mask(self, p: Path) -> np.ndarray:
        mask = Image.open(p)
        return np.asarray(mask)

    @cached_property
    def file_list(self) -> List[BundledPath]:
        root_dir = self.root_dir
        parent = f'BBBC006_v1_images_z_{self.z_ind:02d}'
        _file_list = sorted(root_dir.glob(f'{parent}/*.tif'))
        file_list = bundle_list(_file_list, 2)
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        parent = 'BBBC006_v1_labels'
        anno_list = sorted(root_dir.glob(f'{parent}/*.png'))
        anno_dict = dict((k, v) for k, v in enumerate(anno_list))
        return anno_dict



import xml.etree.ElementTree as ET
from functools import cached_property
from pathlib import Path
from typing import Dict, Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile
from PIL import Image
from skimage.draw import polygon

from ..base import MaskDataset


class ComputationalPathology(MaskDataset):
    """A Dataset and a Technique for Generalized Nuclear Segmentation for
    Computational Pathology [1]_

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ```transforms``` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    mask_tif : bool, default: False
        Instead of parsing every xml file to reconstruct mask image arrays, use
        pre-drawn mask tif files which should reside in the same folder as
        annotation xml files.

    Notes
    -----
    - Resolution of all images is (1000,1000)
    - gt is converted from annotation recorded in xml format
    - gt has dtype of torch.float64, converted from numpy.uint16, and it has
      value 'num_objects' * 255 because it is base-transformed
    - The origianl dataset provides annotation in xml format, which takes
      long time to parse and to reconstruct mask images dynamically during
      training. Drawing masks beforehand makes training much faster. Use
      ``mask_tif`` in that case.
    - When ``augmenters`` is provided, set the ``num_calls`` argument
      30x1000x1000 -> 16x30=480 patches. Thus, the default ``num_calls=720``
      (x1.5)
    - dtype of 'gt' is int16. However, to make batching easier, it will be
      casted to float32
    - Be careful about types of augmenters; avoid interpolation

    References
    ----------
    .. [1] N. Kumar, R. Verma, S. Sharma, S. Bhargava, A. Vahadane, and A.
       Sethi, “A Dataset and a Technique for Generalized Nuclear Segmentation
       for Computational Pathology,” IEEE Transactions on Medical Imaging, vol.
       36, no. 7, pp. 1550–1560, Jul. 2017, doi: 10.1109/TMI.2017.2677499.

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'ComPath'
    # Hard code resolution to parse annotation (.xml)
    _resolution = (1000, 1000)

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # specific to this dataset
        mask_tif: bool = False,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self.mask_tif = mask_tif

    def get_image(self, p: Path) -> np.ndarray:
        img = Image.open(p)
        if img.mode == 'RGBA':
            img = img.convert(mode='RGB')
        return np.asarray(img)

    def get_mask(self, p: Path) -> np.ndarray:
        if self.mask_tif:
            if p.suffix == '.xml':
                raise ValueError(
                    "Use `save_xml_to_tif()` then set `mask_tif` to True"
                )
            mask = tifffile.imread(p)
            return mask.astype(np.int16)
        # Parse xml
        mask = self._parse_xml_to_array(p)
        return mask

    @classmethod
    def _parse_xml_to_array(cls, f_anno) -> np.ndarray:
        """This dataset provides annotation in .xml format

        Consider pre-generating mask image using ``save_xml_to_tif()``
        """
        tree = ET.parse(f_anno)
        root = tree.getroot()

        rr = []
        cc = []
        for region in root.iter('Region'):
            r = []
            c = []
            # print(region.attrib)
            # print(region.find('Vertices'))
            if (vertices := region.find('Vertices')) is not None:
                for v in vertices:
                    # print(v.attrib)
                    r.append(v.attrib['Y'])
                    c.append(v.attrib['X'])
                rr.append(np.array(r, dtype=np.float16))
                cc.append(np.array(c, dtype=np.float16))
        # X, Y = anno['X'], anno['Y']
        mask = np.zeros(cls._resolution, dtype=np.int16)
        for i, (x, y) in enumerate(zip(cc, rr), 1):
            r, c = polygon(y, x, shape=cls._resolution)
            if len(rr) == 0 and len(cc) == 0:
                continue
            mask[r, c] = i
        return mask
        # return {'X': cc, 'Y': rr}

    def save_xml_to_tif(self):
        """Parse .xml to mask and write it as tiff file

        Having masks in images is much faster than parsing .xml for each call.
        This func iterates through ``anno_dict``, parse and save each in .tif
        format in the same annotation directory. Re-initiate an instance with
        ``mask_tif`` argument to load them.
        """
        if self.output not in ['mask', 'both']:
            raise ValueError("Set output either to 'mask' or 'both'")
        for i, p in self.anno_dict.items():
            mask = self._parse_xml_to_array(p)
            fname = p.with_suffix('.tif')
            tifffile.imwrite(
                fname,
                data=mask,
                compression='zlib'
            )
            print(f"[{i}/{len(self.anno_dict) - 1}] Wrote '{fname}'")

    @cached_property
    def file_list(self) -> list:
        root_dir = self.root_dir
        parent = 'Tissue images'
        file_list = sorted(root_dir.glob(f'{parent}/*.png'))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        parent = 'Annotations'
        ext = 'xml'
        if self.mask_tif:
            ext = 'tif'
        anno_dict = dict((k, v) for k, v in enumerate(
            sorted(root_dir.glob(f'{parent}/*.{ext}'))
            ))
        return anno_dict



import os.path
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional

import albumentations
import numpy as np
try:
    from gimpformats.gimpXcfDocument import GimpDocument
except ModuleNotFoundError as e:
    print("Install `gimpformats` pkg")
    raise e

from PIL import Image

from ..base import MaskDataset


class MurphyLab(MaskDataset):
    """Nuclei Segmentation In Microscope Cell Images: A Hand-Segmented Dataset
    And Comparison Of Algorithms [1]_

    Parameters
    ----------
    root_dir : str or pathlib.Path
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines
        augmentation in sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    drop_missing_pairs : bool, default: True
        Valid only if `output='both'`. It will drop images that do not have
        mask pairs.
    drop_broken_files : bool, default: True
        Drop broken files that cannot be read
    filled_mask : bool, default: False
        Use saved filled masks through `fill_save_mask()` method instead of
        default boundary masks. If one would want to use manually modified
        masks, the annotation files should have the same name as '*.xcf'
        with modified suffix by '.png'.

    Warnings
    --------
    This dataset has many issues whose details can be found below. The simpleset
    way is to drop those that cause isseus. It is recommended to not opt out
    ``drop_missing_pairs()`` and ``drop_broken_files()``. Otherwise, it will
    meet exceptions.

    If one wants filled hole, ``fill_save_mask()`` function will fill holes with
    some tricks to handle edge cases and save them as .png format. Then set
    ``filled_mask`` argument to True to load them.

    Read more in Notes section

    Notes
    -----
    - 4 channel PNG format annotation mask even though mask is binary. But it is
      not grayscale binary. They put value 255 only in red channel.
    - Two annotation formats; Photoshop and GIMP. It seems that two annotators
      worked separately. segmented-ashariff will be ignored. In total, 97
      segmented images (out of 100)
    - 3 missing segmentations: ind={31, 43, 75}
        ./data/images/dna-images/gnf/dna-31.png

        ./data/images/dna-images/gnf/dna-43.png

        ./data/images/dna-images/ic100/dna-25.png

    - Manually filled annotation to make masks using GIMP
    - 2009_ISBI_2DNuclei_code_data/data/images/segmented-lpc/ic100/dna-15.xcf
      does not have 'borders' layer like the others.  This one alone has
      'border' layer.

    References
    ----------
    .. [1] L. P. Coelho, A. Shariff, and R. F. Murphy, “Nuclear segmentation in
       microscope cell images: A hand-segmented dataset and comparison of
       algorithms,” in 2009 IEEE International Symposium on Biomedical Imaging:
       From Nano to Macro, Jun. 2009, pp. 518–521, doi:
       10.1109/ISBI.2009.5193098.

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'MurphyLab'

    def __init__(
        self,
        # Interface requirement
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        # Specific to this dataset
        drop_missing_pairs: bool = True,
        drop_broken_files: bool = True,
        filled_mask: bool = False,
        **kwargs
    ):
        # Interface and super-class arguments
        self._root_dir = os.path.join(root_dir, 'data', 'images')
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        # Specific to this dataset
        self.drop_missing_pairs = drop_missing_pairs
        self.drop_broken_files = drop_broken_files
        self.filled_mask = filled_mask

        if self.output == 'both' and self.drop_missing_pairs:
            self.file_list, self.anno_dict = self._drop_missing_pairs()
        if self.output == 'both' and self.drop_broken_files:
            self.file_list, self.anno_dict = self._drop_broken_files()

    def get_image(self, p: Path) -> np.ndarray:
        img = Image.open(p)
        return np.asarray(img)

    def get_mask(self, p: Path) -> np.ndarray:
        if self.filled_mask:
            mask = Image.open(p)
            return np.asarray(mask, dtype=np.float32)
        doc = GimpDocument(p.as_posix())
        layers = [layer.name for layer in doc.layers]
        try:
            ind_layer = layers.index('borders')
        except ValueError:
            # '/data/images/segmented-lpc/ic100/dna-15.xcf' do not have
            # 'borders' but 'border' layer.
            ind_layer = layers.index('border')
        # get image
        layer_borders = doc.getLayer(ind_layer)
        mask = layer_borders.image
        return np.asarray(mask)[..., 0]

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = 'dna-images'
        # dna-images
        file_list = sorted(
            root_dir.glob(f'{parent}/*/*.png'), key=self._sort_key
        )
        return file_list

    @staticmethod
    def _sort_key(p, zfill=2):
        split = p.stem.split('-')
        return '-'.join([p.parent.stem] + [s.zfill(zfill) for s in split])

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        ext = '.png' if self.filled_mask else '.xcf'
        anno_dict = {}
        for i, p in enumerate(self.file_list):
            p_anno = '/'.join([p.parent.stem, p.stem + ext])
            # Ignore 'segmented-ashariff`. It seems that Ashariff got bored
            # after 10 images.
            anno = p.parents[2] / 'segmented-lpc' / p_anno
            if anno.exists():
                anno_dict[i] = anno
        return anno_dict

    def _drop_broken_files(self):
        """Drop broken files

        '/data/images/segmented-lpc/ic100/dna-46.xcf' cannnot be read by
        ``gimpformats``
        """
        file_list = self.file_list
        anno_dict = self.anno_dict
        for i, p in anno_dict.items():
            # if p.name == 'dna-46.xcf':
            if '/'.join([p.parent.name, p.name]) == 'ic100/dna-46.xcf':
                file_list.pop(i)
                anno_dict.pop(i)
                break
        anno_dict = dict((i, v) for i, v in enumerate(anno_dict.values()))
        return file_list, anno_dict

    def fill_save_mask(self):
        """Fill holes from boundary mask with some tricks

        Requires scipy and scikit-image. Install depencency with pip option
        ``pip install bioimageloader[process]``.

        Note that this does not result perfect filled masks. Those not entirely
        closed by this algorithm (36, 40, 63).

        Other issues: ``ind=63``: 'border' not 'borders', ``ind=93``
        ``GimpDocument`` cannot read it...
        """
        from scipy.ndimage import binary_fill_holes
        from skimage.morphology import dilation, erosion

        def fill_holes(
            img: np.ndarray,
            w_edge: int,
            w_pad: int
        ) -> np.ndarray:
            # cut edges because many are not closed
            edge_cut = img[w_edge:-w_edge, w_edge:-w_edge]
            # dilate to connect some boundaries
            dilated = dilation(edge_cut)
            # pad with refletion mode to close bounary at edges
            filled_pad = binary_fill_holes(
                np.pad(dilated, w_pad, mode='reflect')
            )
            # back to original shape
            w_rev = w_pad - w_edge
            filled_pad = filled_pad[w_rev:-w_rev, w_rev:-w_rev]
            # erode, because we dilated
            filled_pad = erosion(filled_pad)
            return filled_pad

        for k, p in self.anno_dict.items():
            # read gimp document .xcf
            try:
                doc = GimpDocument(p.as_posix())
            except TypeError:
                # '/data/images/segmented-lpc/ic100/dna-46.xcf' cannnot be read
                # through gimpformats library. You can open it GIMP though.
                print(f"Cannot open '{k}: {p}' with gimpformats lib")
                continue
            # get 'borders' layer
            layers = [layer.name for layer in doc.layers]
            try:
                ind_layer = layers.index('borders')
            except ValueError:
                # '/data/images/segmented-lpc/ic100/dna-15.xcf' do not have
                # 'borders' but 'border' layer.
                print(f"Exception layer['border'] '{k}: {p}'")
                ind_layer = layers.index('border')
            # get image
            layer_borders = doc.getLayer(ind_layer)
            mask = layer_borders.image
            mask = np.asarray(mask)
            # print(k, mask.shape)
            # fill_holes, it has 3 channels and red channel has annotation
            filled = fill_holes(mask[..., 0], w_edge=15, w_pad=200)
            # save
            img = Image.fromarray(filled)
            print(f"Saving '{k}: {p}'")
            img.save(p.with_suffix('.png'))



"""Classic utils module
"""

import csv
import random
from copy import deepcopy
from itertools import accumulate
from pathlib import Path
from typing import Callable, List, Optional, Protocol, Sequence, TypeVar, Union

import albumentations
import numpy as np
from PIL import Image

from .base import Dataset, MaskDataset
from .types import Bundled

T = TypeVar('T')


class MaskDatasetProto(Protocol):
    """Static typing protocol for MaskDataset
    """
    file_list: list
    anno_dict: dict
    output: str

    def __len__(self):
        ...


def random_label_cmap(n=2**16, h=(0, 1), l=(.4, 1), s=(.2, .8)):
    """Random color map for labels (credit: StarDist team) [1]_

    Need matplotlib

    .. [1] https://github.com/stardist/stardist/blob/4422d1c235175a41d657009cb01075347cd14a53/stardist/plot/plot.py#L8
    """
    import colorsys

    import matplotlib

    # cols = np.random.rand(n,3)
    # cols = np.random.uniform(0.1,1.0,(n,3))
    h,l,s = np.random.uniform(*h,n), np.random.uniform(*l,n), np.random.uniform(*s,n)
    cols = np.stack([colorsys.hls_to_rgb(_h,_l,_s) for _h,_l,_s in zip(h,l,s)],axis=0)
    cols[0] = 0
    return matplotlib.colors.ListedColormap(cols)


def imread_asarray(p: Path, dtype=None) -> np.ndarray:
    '''Read an image using PIL then convert it into numpy array'''
    img = np.asarray(Image.open(p), dtype=dtype)
    return img


def read_csv(file: Union[str, Path], sniffer_siz=2048) -> tuple:
    with open(file, newline='') as csvfile:
        sniffer = csv.Sniffer()
        dialect = sniffer.sniff(csvfile.readline())
        has_header = sniffer.has_header(csvfile.readline())
        csvfile.seek(0)
        reader = csv.reader(csvfile, dialect)
        header = None
        if has_header:
            header = next(reader)
        lines = [row for row in reader]
    return header, lines


def ordered_unique(seq: Sequence[T]) -> List[T]:
    unique = []
    v = None
    for _v in seq:
        if v != _v:
            unique.append(_v)
            v = _v
    return unique


def rle_decoding_inseg(
    size: Union[tuple, list],
    run_lengths: List[List[int]],
) -> np.ndarray:
    """Decoding RLE (Run Length Encoding). Output binary mask. If you want each
    instance have different values, use `rle_decoding_inseg(), instead.`

    Parameters
    ----------
    size : list or tuple
        Shape of the original image array (height, width)
    run_lengths : list
        List of run length encodings
    val : int or float
        Constant value for all encoded pixels

    Returns
    -------
    decoded : numpy.ndarray
        Decoded image array

    """
    # #--- Draw canvas ---# #
    h, w = size[0], size[1]
    decoded = np.zeros(h * w, dtype=np.uint8)
    # num_objs = len(run_lengths)

    for i, rle in enumerate(run_lengths):
        for p, l in zip(rle[0::2], rle[1::2]):
            for dp in range(l):
                decoded[(p-1)+dp] = i + 1  # 0 is bg
    return decoded.reshape((w, h)).T


def subset(dataset: MaskDatasetProto, indices: Sequence[int]):
    indices = sorted(indices)
    dset = deepcopy(dataset)
    dset.file_list = [dataset.file_list[i] for i in indices]
    if dataset.output != 'image':
        dset.anno_dict = dict((i, dataset.anno_dict[k]) for i, k in enumerate(
            indices
        ))
    return dset


def random_split_dataset(
    dataset: MaskDatasetProto,
    lengths: Sequence[int],
) -> List[MaskDataset]:
    """Randomly split dataset and return subsets
    """
    if sum(lengths) != len(dataset):
        raise ValueError("Sum of input lengths does not equal the length of the input dataset!")
    indices = list(range(len(dataset)))
    random.shuffle(indices)
    return [subset(dataset, indices[offset - length:offset])
            for offset, length in zip(accumulate(lengths), lengths)]


def split_dataset_by_indices(
    dataset: MaskDatasetProto,
    indices: Sequence[int],
) -> List[MaskDataset]:
    """Split dataset given indices
    """
    return subset(dataset, indices)


def stack_channels(
    imread_handler: Callable[[Path], np.ndarray],
    p_lst: List[Path],
    *axis_order: int
) -> np.ndarray:
    """Take a list of multi-channel images whose channels are separated in each
    file and read them in specified order.

    The order of channels follows the order of each list by default. If
    `*axis_order` is explicitely given, the function will put channels in that
    order.

    Parameters
    ----------
    imread_handler : Callable
        Func to read images e.g.) PIL.Image.open
    p_lst : list of file path
        A list of file path. Each element refers to one channel.
    axis_order : int(s)
        Additional arguments to indicate the order of channels. It should match
        the number of channels of the return. For example, 3 arguments if
        num_channels <= 3, else `n` arguments elif num_channels=`n`

    """
    images = []
    for p in p_lst:
        images.append(imread_handler(p))
    num_channels = len(images)
    if (num_axes := len(axis_order)) != 0:
        if num_channels != num_axes:
            raise ValueError
    stacked = np.stack(images, axis=-1)
    if axis_order:
        ordered = np.zeros_like(stacked)
        for i, o in enumerate(axis_order):
            ordered[..., o] = stacked[..., i]
        return ordered
    return stacked


def stack_channels_to_rgb(
    imread_handler: Callable[[Path], np.ndarray],
    p_lst: List[Path],
    *axis_order: int
) -> np.ndarray:
    """Many transforms work for either RGB or gray scale images. Having RGB is
    also helpful for visualization.

    Take a list of multi-channel images whose channels are separated in each
    file and read them in specified order. If the number of channels is less
    than or equal to 3, then array will be assumed as a RGB image. Otherwise, it
    it returns an array with the same number of channels of the input.

    The order of channels follows the order of each list by default. If
    `*axis_order` is explicitely given, the function will put channels in that
    order.

    Parameters
    ----------
    imread_handler : Callable
        Func to read images e.g.) PIL.Image.open | tifffile.imread
    p_lst : a list of Paths
        A list of Path objects. Each element refers to one channel
    axis_order : int(s)
        Additional arguments to indicate the order of channels. It should match
        the number of channels of the return. For example, 3 arguments if
        num_channels <= 3, else `n` arguments elif num_channels=`n`

    """
    images = []
    for p in p_lst:
        images.append(imread_handler(p))
    num_channels = len(images)
    stacked = np.stack(images, axis=-1)
    if num_channels < 3:
        # it happens to be only 2 channels
        stacked = np.concatenate([stacked, np.zeros_like(images[0])[..., np.newaxis]], axis=-1)
    if axis_order:
        ordered = np.zeros_like(stacked)
        for i, o in enumerate(axis_order):
            ordered[..., o] = stacked[..., i]
        return ordered
    return stacked


def bundle_list(lst: List[T], bundle_size: int) -> List[Bundled[T]]:
    """Reshape a list given the repetition step size"""
    return [list(e) for e in zip(
        *[lst[i::bundle_size] for i in range(bundle_size)]
    )]


def expand_to_rgb(
    image: np.ndarray,
    dtype: Optional[str] = None,
) -> np.ndarray:
    """Expand axis of image that has 2 channels to have 3 channels mainly for
    visualization
    """
    num_channels = image.shape[-1]
    if num_channels != 2:
        raise ValueError
    # it happens to be only 2 channels
    stacked = np.concatenate(
        [image, np.zeros_like(image[..., 0])[..., np.newaxis]],
        axis=-1,
    )
    if dtype is not None:
        return stacked.astype(dtype, copy=False)
    return stacked


def get_dataset_from_directory(
    root_dir: str,
    *,
    output: Optional[str] = None,
    transforms: Optional[albumentations.Compose] = None,
    num_calls: Optional[int] = None,
    grayscale: Optional[bool] = None,
    grayscale_mode: Optional[Union[str, Sequence[float]]] = None,
) -> Dataset:
    """Construct MaskDataset by assuming the structure of given directory

    >>> case1/
    ├── image00.tif
    ├── image01.tif
    ├── image02.tif
    ├── image03.tif
    ├── image04.tif
    ├── image05.tif
    ├── image06.tif
    ├── image07.tif
    ├── image08.tif
    └── image09.tif

    """
    # works with case1
    # case1/                  *case2/                *case4/
    # ├── image00.tif         ├── image00.tif         ├── images
    # ├── image01.tif         ├── image01.tif         │   ├── 00.png
    # ├── image02.tif         ├── image02.tif         │   ├── 01.png
    # ├── image03.tif         ├── image03.tif         │   ├── 02.png
    # ├── image04.tif         ├── image04.tif         │   ├── 03.png
    # ├── image05.tif         ├── label00.tif         │   └── 04.png
    # ├── image06.tif         ├── label01.tif         └── labels
    # ├── image07.tif         ├── label02.tif             ├── 00
    # ├── image08.tif         ├── label03.tif             │   ├── 0.jpg
    # └── image09.tif         └── label04.tif             │   ├── 1.jpg
    #                                                     │   ├── 3.jpg
    # case3/                                              │   └── 4.jpg
    # ├── images                                          ├── 01
    # │   ├── 00.png                                      │   ├── 0.jpg
    # │   ├── 01.png                                      │   ├── 1.jpg
    # │   ├── 02.png                                      │   ├── 2.jpg
    # │   ├── 03.png                                      │   ├── 3.jpg
    # │   ├── 04.png                                      │   ├── 4.jpg
    # │   ├── 05.png                                      │   ├── 5.jpg
    # │   ├── 06.png                                      │   └── 6.jpg
    # │   ├── 07.png                                      ├── 02
    # │   ├── 08.png                                      │   ├── 0.jpg
    # │   └── 09.png                                      │   ├── 1.jpg
    # └── labels                                          │   └── 2.jpg
    #     ├── 00.tif                                      ├── 03
    #     ├── 01.tif                                      │   ├── 0.jpg
    #     ├── 02.tif                                      │   ├── 1.jpg
    #     ├── 03.tif                                      │   ├── 2.jpg
    #     ├── 04.tif                                      │   ├── 3.jpg
    #     ├── 05.tif                                      │   ├── 4.jpg
    #     ├── 06.tif                                      │   └── 5.jpg
    #     ├── 07.tif                                      └── 04
    #     ├── 08.tif                                          ├── 0.jpg
    #     └── 09.tif                                          └── 1.jpg

    from .common import CommonDataset
    mask_dataset = CommonDataset(
        root_dir=root_dir,
        output=output,
        transforms=transforms,
        num_calls=num_calls,
        grayscale=grayscale,
        grayscale_mode=grayscale_mode,
    )
    return mask_dataset


def get_maskdataset_from_directory(
    root_dir: str,
    *,
    image_dir: Optional[str] = None,
    label_dir: Optional[str] = None,
    output: Optional[str] = None,
    transforms: Optional[albumentations.Compose] = None,
    num_calls: Optional[int] = None,
    grayscale: Optional[bool] = None,
    grayscale_mode: Optional[Union[str, Sequence[float]]] = None,
) -> MaskDataset:
    """Construct MaskDataset by assuming the structure of given directory

    >>> case3/
    ├── images
    │   ├── 00.png
    │   ├── 01.png
    │   ├── 02.png
    │   ├── 03.png
    │   └── 04.png
    └── labels
        ├── 00.tif
        ├── 01.tif
        ├── 02.tif
        ├── 03.tif
        └── 04.tif

    """
    # work with case3
    # case1/                  *case2/                *case4/
    # ├── image00.tif         ├── image00.tif         ├── images
    # ├── image01.tif         ├── image01.tif         │   ├── 00.png
    # ├── image02.tif         ├── image02.tif         │   ├── 01.png
    # ├── image03.tif         ├── image03.tif         │   ├── 02.png
    # ├── image04.tif         ├── image04.tif         │   ├── 03.png
    # ├── image05.tif         ├── label00.tif         │   └── 04.png
    # ├── image06.tif         ├── label01.tif         └── labels
    # ├── image07.tif         ├── label02.tif             ├── 00
    # ├── image08.tif         ├── label03.tif             │   ├── 0.jpg
    # └── image09.tif         └── label04.tif             │   ├── 1.jpg
    #                                                     │   ├── 3.jpg
    # case3/                                              │   └── 4.jpg
    # ├── images                                          ├── 01
    # │   ├── 00.png                                      │   ├── 0.jpg
    # │   ├── 01.png                                      │   ├── 1.jpg
    # │   ├── 02.png                                      │   ├── 2.jpg
    # │   ├── 03.png                                      │   ├── 3.jpg
    # │   ├── 04.png                                      │   ├── 4.jpg
    # │   ├── 05.png                                      │   ├── 5.jpg
    # │   ├── 06.png                                      │   └── 6.jpg
    # │   ├── 07.png                                      ├── 02
    # │   ├── 08.png                                      │   ├── 0.jpg
    # │   └── 09.png                                      │   ├── 1.jpg
    # └── labels                                          │   └── 2.jpg
    #     ├── 00.tif                                      ├── 03
    #     ├── 01.tif                                      │   ├── 0.jpg
    #     ├── 02.tif                                      │   ├── 1.jpg
    #     ├── 03.tif                                      │   ├── 2.jpg
    #     ├── 04.tif                                      │   ├── 3.jpg
    #     ├── 05.tif                                      │   ├── 4.jpg
    #     ├── 06.tif                                      │   └── 5.jpg
    #     ├── 07.tif                                      └── 04
    #     ├── 08.tif                                          ├── 0.jpg
    #     └── 09.tif                                          └── 1.jpg

    from .common import CommonMaskDataset
    mask_dataset = CommonMaskDataset(
        root_dir=root_dir,
        output=output,
        transforms=transforms,
        num_calls=num_calls,
        grayscale=grayscale,
        grayscale_mode=grayscale_mode,
    )
    mask_dataset._setattr_ifvalue('_image_dir', image_dir)
    mask_dataset._setattr_ifvalue('_label_dir', label_dir)

    return mask_dataset



from pathlib import Path
from typing import Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile

from bioimageloader.collections import UCSB


class UCSBNUnet(UCSB):
    """A biosegmentation benchmark for evaluation of bioimage analysis methods
    (Adopted for NU-Net)

    Notes
    -----
    - 32 'benign', 26 'malignant' images (58 images in total)
    - 58x768x896 -> ~600 patches. Thus, the defulat `num_calls=900` (x1.5).
    - Images are not fully annotated

    References
    ----------
    .. [1] E. Drelie Gelasca, B. Obara, D. Fedorov, K. Kvilekval, and B.
       Manjunath, “A biosegmentation benchmark for evaluation of bioimage
       analysis methods,” BMC Bioinformatics, vol. 10, p. 368, Nov. 2009, doi:
       10.1186/1471-2105-10-368.
    """
    # Dataset's acronym
    acronym = 'UCSB'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # specific to this dataset
        category: Sequence[str] = ('malignant',),
        # NU-Net args
        invt: bool = True,
        **kwargs
    ):
        """
        Parameters
        ----------
        root_dir : str
            Path to root directory
        output : {'image','mask','both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        grayscale : bool (default: False)
            Convert images to grayscale
        grayscale_mode : {'cv2', 'equal', Sequence[float]} (default: 'cv2')
            How to convert to grayscale. If set to 'cv2', it follows opencv
            implementation. Else if set to 'equal', it sums up values along
            channel axis, then divides it by the number of expected channels.
        category : {'benign', 'malignant'} (default: ('malignant',))
            Select which category of output you want
        invt : bool (default: True)
            Invert image pixel values

        See Also
        --------
        UCSB : Super class
        MaskDataset : Super class
        DatasetInterface : Interface

        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            grayscale=grayscale,
            grayscale_mode=grayscale_mode,
            category=category,
            **kwargs
        )
        self.invt = invt

    def get_image(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        return ~tif



from functools import cached_property

from ..base import MaskDataset


class Minimal(MaskDataset):

    acronym = 'Minimal'

    def __init__(self, root_dir):
        self._root_dir = root_dir

    @cached_property
    def file_list(self) -> list:
        return list('abc')

    def get_image(self):
        ...



from functools import cached_property
from pathlib import Path
from typing import Dict, Optional

import albumentations
import numpy as np
from PIL import Image

from bioimageloader.collections import MurphyLab


class MurphyLabNUnet(MurphyLab):
    """Nuclei Segmentation In Microscope Cell Images: A Hand-Segmented Dataset
    And Comparison Of Algorithms (Adopted for NU-Net)

    Mask variation. Manually filled boundary labels with a bucket tool and saved
    in .png format.


    IMPORTANT
    ---------
    This dataset has many issues whose details can be found below. The simpleset
    way is to drop those that cause isseus. It is recommended to not opt out
    `drop_missing_pairs()` and `drop_broken_files()`. Otherwise, it will meet
    exceptions.

    If one wants filled hole, `fill_save_mask()` function will fill holes with
    some tricks to handle edge cases and save them as .png format. Then set
    `filled_mask` argument to True to load them.


    Notes
    -----
    - Two annotation formats; Photoshop and GIMP. It seems that two annotators
      worked separately. segmented-ashariff will be ignored. In total, 97
      segmented images (out of 100)
    - 3 missing segmentations: ind={31, 43, 75}
        ./data/images/dna-images/gnf/dna-31.png
        ./data/images/dna-images/gnf/dna-43.png
        ./data/images/dna-images/ic100/dna-25.png
    - Manually filled annotation to make masks using GIMP
    - 2009_ISBI_2DNuclei_code_data/data/images/segmented-lpc/ic100/dna-15.xcf
      does not have 'borders' layer like the others.  This one alone has
      'border' layer.

    .. [1] L. P. Coelho, A. Shariff, and R. F. Murphy, “Nuclear segmentation in
       microscope cell images: A hand-segmented dataset and comparison of
       algorithms,” in 2009 IEEE International Symposium on Biomedical Imaging:
       From Nano to Macro, Jun. 2009, pp. 518–521, doi:
       10.1109/ISBI.2009.5193098.
    """

    # Dataset's acronym
    acronym = 'MurphyLab'

    def __init__(
        self,
        # Interface requirement
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        # Specific to this dataset
        drop_missing_pairs: bool = True,
        drop_broken_files: bool = False,
        filled_mask: bool = True,
        **kwargs
    ):
        """
        Parameters
        ---------
        root_dir : str or pathlib.Path
            Path to root directory
        output : {'image','mask','both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        drop_missing_pairs : bool (default: True)
            Valid only if `output='both'`. It will drop images that do not have
            mask pairs.
        drop_broken_files : bool (default: False)
            Drop broken files that cannot be read
        filled_mask : bool (default: True)
            Use saved filled masks through `fill_save_mask()` method instead of
            default boundary masks. If one would want to use manually modified
            masks, the annotation files should have the same name as '*.xcf'
            with modified suffix by '.png'.

        See Also
        --------
        MurphyLab : Super class
        MaskDataset : Super class
        DatasetInterface : Interface

        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            drop_missing_pairs=drop_missing_pairs,
            drop_broken_files=drop_broken_files,
            filled_mask=filled_mask,
            **kwargs
        )

    def get_mask(self, p: Path) -> np.ndarray:
        mask = Image.open(p)
        # Red channel is where anno is
        return np.asarray(mask)[..., 0]

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        ext = '.png' if self.filled_mask else '.xcf'
        anno_dict = {}
        for i, p in enumerate(self.file_list):
            p_anno = '/'.join([p.parent.stem, p.stem + '-filled' + ext])
            # Ignore 'segmented-ashariff`. It seems that Ashariff got bored
            # after 10 images.
            anno = p.parents[2] / 'segmented-lpc' / p_anno
            if anno.exists():
                anno_dict[i] = anno
        return anno_dict



from functools import cached_property
from pathlib import Path
from typing import List, Optional

import albumentations
import cv2
import numpy as np

from ..base import Dataset
from ..utils import imread_asarray


class BBBC026(Dataset):
    """Human Hepatocyte and Murine Fibroblast cells – Co-culture experiment

    This 384-well plate has images of co-cultured hepatocytes and fibroblasts.
    Every other well is populated (A01, A03, ..., C01, C03, ...) such that 96
    wells comprise the data. Each well has 9 sites and thus 9 images associated,
    totaling 864 images.

    For each well there is one field and a single image nuclear image (Hoecsht).
    Images are in 8-bit PNG format.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.

    Notes
    -----
    - Only centers are annotated for 5 imgages (not implemented)

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC026

    See Also
    --------
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC026'

    def __init__(
        self,
        root_dir: str,
        *,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_calls = num_calls

    def get_image(self, p: Path) -> np.ndarray:
        img = imread_asarray(p)
        return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = 'BBBC026_v1_images'
        file_list = sorted(root_dir.glob(f'{parent}/*.png'))
        return file_list



import os.path
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile
from PIL import Image

from bioimageloader.collections import DigitalPathology


class DigitalPathologyNUnet(DigitalPathology):
    """Deep learning for digital pathology image analysis: A comprehensive
    tutorial with selected use cases (Adopted for NU-Net)

    Notes
    -----
    - Annotation is partial, thus `output='image'` by default.

    References
    ----------
    .. [1] A. Janowczyk and A. Madabhushi, “Deep learning for digital pathology
       image analysis: A comprehensive tutorial with selected use cases,” J
       Pathol Inform, vol. 7, Jul. 2016, doi: 10.4103/2153-3539.186902.
    """
    # Dataset's acronym
    acronym = 'DigitPath'

    def __init__(
        self,
        # Interface requirement
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # NU-Net args
        invt: bool = True,
        **kwargs
    ):
        """
        Parameters
        ----------
        root_dir : str
            Path to root directory
        output : {'image','mask','both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        grayscale : bool (default: False)
            Convert images to grayscale
        grayscale_mode : {'cv2', 'equal', Sequence[float]} (default: 'cv2')
            How to convert to grayscale. If set to 'cv2', it follows opencv
            implementation. Else if set to 'equal', it sums up values along
            channel axis, then divides it by the number of expected channels.
        invt : bool (default: True)
            Invert image pixel values

        See Also
        --------
        DigitalPathology : Super class
        MaskDataset : Super class
        DatasetInterface : Interface

        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            grayscale=grayscale,
            grayscale_mode=grayscale_mode,
            **kwargs
        )
        self.invt = invt

    def get_image(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        if self.invt:
            return ~tif
        return tif



import os.path
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np
from PIL import Image

from ..base import MaskDataset


class TNBC(MaskDataset):
    """TNBC Nuclei Segmentation Dataset [1]_

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines
        augmentation in sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along
        channel axis, then divides it by the number of expected channels.

    References
    ----------

    .. [1] Segmentation of Nuclei in Histopathology Images by Deep Regression of
       the Distance Map, https://ieeexplore.ieee.org/document/8438559

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface
    """

    # Dataset's acronym
    acronym = 'TNBC'

    def __init__(
        self,
        # Interface requirement
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        **kwargs
    ):
        # Interface and super-class arguments
        self._root_dir = os.path.join(root_dir, 'TNBC_NucleiSegmentation')
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode

    def get_image(self, p: Path) -> np.ndarray:
        img = Image.open(p)
        if img.mode == 'RGBA':
            img = img.convert(mode='RGB')
        return np.asarray(img)

    def get_mask(self, p: Path) -> np.ndarray:
        mask = Image.open(p)
        return np.asarray(mask)

    @cached_property
    def file_list(self) -> List[Path]:
        # Call MaskDataset.root_dir
        root_dir = self.root_dir
        parent = 'Slide_*'
        file_list = sorted(root_dir.glob(f'{parent}/*.png'))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        """anno_dict[ind] = <file>"""
        root_dir = self.root_dir
        parent = 'GT_*'
        anno_dict = dict((k, v) for k, v in enumerate(
            sorted(root_dir.glob(f'{parent}/*.png'))
            ))
        return anno_dict



import re
from functools import cached_property, partial
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile

from ..base import MaskDataset


class UCSB(MaskDataset):
    """A biosegmentation benchmark for evaluation of bioimage analysis methods

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both',' image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines
        augmentation in sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along
        channel axis, then divides it by the number of expected channels.
    category : {'benign', 'malignant'}, default: ('malignant',)
        Select which category of output you want

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    Notes
    -----
    - 32 'benign', 26 'malignant' images (58 images in total)
    - 58x768x896 -> ~600 patches. Thus, the defulat `num_calls=900` (x1.5).
    - Images are not fully annotated

    References
    ----------
    .. [1] E. Drelie Gelasca, B. Obara, D. Fedorov, K. Kvilekval, and B.
       Manjunath, “A biosegmentation benchmark for evaluation of bioimage
       analysis methods,” BMC Bioinformatics, vol. 10, p. 368, Nov. 2009, doi:
       10.1186/1471-2105-10-368.
    """
    # Dataset's acronym
    acronym = 'UCSB'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # specific to this dataset
        category: Sequence[str] = ('malignant',),
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self.category = category
        if not any([cat in ('benign', 'malignant') for cat in category]):
            raise ValueError("Set `category` in ('benign', 'malignant') in sequence")

    def get_image(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        return tif

    def get_mask(self, p: Path) -> np.ndarray:
        tif = tifffile.imread(p)
        return tif

    @staticmethod
    def _filter_category(p: Path, category: str):
        return re.search(category, p.stem)

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = 'Breast Cancer Cells'
        file_list = root_dir.glob(f'{parent}/*.tif')
        if len(cat := self.category) == 1:
            file_list = filter(partial(self._filter_category, category=cat[0]),
                               file_list)
        return sorted(file_list)

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        parent = 'Breast Cancer Cells GroundTruth'
        anno_list = sorted(root_dir.glob(f'{parent}/*.TIF'))
        if len(cat := self.category) == 1:
            anno_list = filter(partial(self._filter_category, category=cat[0]),
                               anno_list)
        anno_dict = dict((k, v) for k, v in enumerate(anno_list))
        return anno_dict



from typing import List, Dict
from .base import Dataset
from .collections import *
from .collections import __all__ as ALL_COLLECTIONS


ROOTS = {
    # Mask anno
    #   Instance
    'DSB2018'                 : '../Data/DSB2018',
    'ComputationalPathology'  : '../Data/ComputationalPathology',
    'FRUNet'                  : '../Data/FRU_processing',
    'S_BSST265'               : '../Data/BioStudies',
    'BBBC006'                 : '../Data/bbbc/006',
    'BBBC020'                 : '../Data/bbbc/020',
    'BBBC039'                 : '../Data/bbbc/039',
    #   Semantic (fg/bg)
    'TNBC'                    : '../Data/TNBC_NucleiSegmentation',
    'MurphyLab'               : '../Data/2009_ISBI_2DNuclei_code_data',
    'BBBC007'                 : '../Data/bbbc/007',
    'BBBC008'                 : '../Data/bbbc/008',
    'BBBC018'                 : '../Data/bbbc/018',
    # Partial mask anno
    'DigitalPathology'        : '../Data/DigitalPathology',
    'UCSB'                    : '../Data/UCSB_BioSegmentation',
    # no anno
    'BBBC002'                 : '../Data/bbbc/002',
    'BBBC013'                 : '../Data/bbbc/013',
    'BBBC014'                 : '../Data/bbbc/014',
    'BBBC015'                 : '../Data/bbbc/015',
    'BBBC016'                 : '../Data/bbbc/016',
    'BBBC026'                 : '../Data/bbbc/026',
    'BBBC041'                 : '../Data/bbbc/041/malaria',  # malaria
    'BBBC021'                 : '../Data/bbbc/021',  # huge number 132,000
}


def load_all_datasets(roots: Dict[str, str] = ROOTS) -> List[Dataset]:
    """Load all available datasets

    You may not want to use it. I use it to take samples from each one for
    visuaization.

    """
    datasets: List[Dataset] = []
    for dataset in ALL_COLLECTIONS:
        exec(f'datasets.append({dataset}("{roots[dataset]}"))')
    return datasets



from pathlib import Path
from typing import Optional, Sequence, Union

import albumentations
import numpy as np
from PIL import Image

from bioimageloader.collections import TNBC


class TNBCNUnet(TNBC):
    """TNBC Nuclei Segmentation Dataset (Adopted for NU-Net)

    References
    ----------
    Segmentation of Nuclei in Histopathology Images by Deep Regression of
    the Distance Map [1]_

    .. [1] https://ieeexplore.ieee.org/document/8438559

    """

    # Dataset's acronym
    acronym = 'TNBC'

    def __init__(
        self,
        # Interface requirement
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = True,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # NU-Net args
        invt: bool = True,
        **kwargs
    ):
        """
        Parameters
        ---------
        root_dir : str
            Path to root directory
        output : {'image','mask','both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        grayscale : bool (default: False)
            Convert images to grayscale
        grayscale_mode : {'cv2', 'equal', Sequence[float]} (default: 'cv2')
            How to convert to grayscale. If set to 'cv2', it follows opencv
            implementation. Else if set to 'equal', it sums up values along
            channel axis, then divides it by the number of expected channels.
        invt : bool (default: True)
            Invert image pixel values

        See Also
        --------
        MaskDataset : Super class
        DatasetInterface : Interface
        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            grayscale=grayscale,
            grayscale_mode=grayscale_mode,
            **kwargs
        )
        self.invt = invt

    def get_image(self, p: Path) -> np.ndarray:
        img = Image.open(p)
        if img.mode == 'RGBA':
            img = img.convert(mode='RGB')
        if self.invt:
            return ~np.asarray(img)
        return np.asarray(img)



from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union, Any

import albumentations
import numpy as np
from PIL import Image

from ..base import MaskDataset
from ..types import BundledPath
from ..utils import imread_asarray, rle_decoding_inseg, read_csv, ordered_unique


class DSB2018(MaskDataset):
    """Data Science Bowl 2018 [1]_ also known as BBBC038 [2]_

    Find the nuclei in divergent images to advance medical discovery

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    training : bool, default: True
        Load training set if True, else load testing one

    References
    ----------
    .. [1] https://www.kaggle.com/c/data-science-bowl-2018/
    .. [2] https://bbbc.broadinstitute.org/BBBC038/

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = 'DSB2018'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # specific to this dataset
        training: bool = True,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this one here
        self.training = training

    def get_image(self, p: Path) -> np.ndarray:
        img = Image.open(p)
        img = img.convert(mode='RGB')
        return np.asarray(img)

    def get_mask(self, anno: Union[BundledPath, Dict[str, Any]]) -> np.ndarray:
        if self.training and not isinstance(anno, dict):
            # anno: BundlePath
            p = anno[0]
            val = 1
            m0 = imread_asarray(p) > 0
            mask = np.zeros_like(m0, dtype=np.uint8)  # uint8 is enough
            mask[m0] = val
            for p in anno[1:]:
                val += 1
                m = imread_asarray(p) > 0
                # Does not allow overlapping!
                mask[m] = val
            return mask
        # anno: dict
        run_lengths = anno['EncodedPixels']
        h, w = anno['Height'], anno['Width']
        mask = rle_decoding_inseg((h, w), run_lengths)
        return mask

    @cached_property
    def ids(self) -> List[str]:
        if self.training:
            _, lines = read_csv(self.root_dir / 'stage1_train_labels.csv')
        else:
            _, lines = read_csv(self.root_dir / 'stage1_solution.csv')
        ids = [line[0] for line in lines]
        ids = ordered_unique(ids)
        return ids

    @cached_property
    def file_list(self) -> List[Path]:
        # Call MaskDataset.root_dir
        parent = 'stage1_train' if self.training else 'stage1_test'
        return [self.root_dir / parent / i / 'images' / f'{i}.png'
                for i in self.ids]

    @cached_property
    def anno_dict(self) -> Union[Dict[int, BundledPath], Dict[int, dict]]:
        if self.training:
            anno_dict = {}
            for i, p in enumerate(self.file_list):
                anno_dict[i] = list(p.parents[1].glob('masks/*.png'))
            return anno_dict
        else:
            anno_rle = {}
            _, lines = read_csv(self.root_dir / 'stage1_solution.csv')
            # header: ImageId,EncodedPixels,Height,Width,Usage
            # iter_rle = map(lambda line: [int(s) for s in line[1].split(' ')],
            #                lines)
            offset = 0
            for i, idx in enumerate(self.ids):
                solution: dict = {'EncodedPixels': []}
                for line in lines[offset:]:
                    if idx == line[0]:
                        if 'Height' not in solution:
                            solution['Height'] = int(line[2])
                            solution['Width'] = int(line[3])
                        solution['EncodedPixels'].append(
                            [int(s) for s in line[1].split(' ')]
                        )
                        offset += 1
                    else:
                        break
                anno_rle[i] = solution
            return anno_rle
